"""A Generic factory implementation that can generate and work with any MARTe1ConfigObject
 or inheriting object.

Place thing function in each python file of something that can be loaded into the Factory:

def initialize(factory, plugin_datastore) -> None:
    factory.registerBlock("{class name to be used}", {class in file to use}, plugin_datastore)

"""

import importlib
import json

class Factory:
    ''' The general factory method, should work with any object that in it's python file
    contains an initialize method which calls and uses the register block. '''
    def __init__(self):
        self.classes = {}

    def importModule(self, name: str):
        ''' Imports a python file '''
        return importlib.import_module(name) # type: ignore

    def unloadAll(self):
        ''' Reset our loaded class instances '''
        self.classes = {}

    def loadRemote(self, path):
        ''' Load from a given JSON definition file '''
        with open(path, 'r', encoding='UTF-8') as json_file:
            plugins = json.load(json_file).values()
        self.loadPlugins(plugins, self.classes)

    def loadPlugins(self, plugins, plugin_datastore) -> None:
        """Load the plugins defined in the plugins list."""
        for plugin_name in plugins:
            plugin = self.importModule(plugin_name)
            plugin.initialize(self, plugin_datastore)

    def registerBlock(self, class_name: str, block_class: object, plugin_datastore):
        ''' Register a class instance into our plugin datastore - typically self.classes '''
        plugin_datastore[class_name] = block_class

    def unregisterBlock(self, class_name: str):
        ''' Unregister a given class instance from our list '''
        try:
            self.classes.pop(class_name)
        except KeyError:
            raise ValueError(f"Unknown block type {class_name}") # pylint: disable=W0707

    def create(self, block_type = "and"):
        """Create a block of a specific type, given an initblock."""
        try:
            return self.classes[block_type]
        except KeyError:
            raise ValueError(f"Unknown block type {block_type}") # pylint: disable=W0707

    def getAll(self):
        ''' Get an instance of all registered block classes. '''
        return list(self.classes.values())



''' Easy import of the majority of objects all at once '''

from martepy.marte2.objects.message import MARTe2Message
from martepy.marte2.objects.http import (MARTe2HttpDirectoryResource,
                                         MARTe2HttpService,
                                         MARTe2HttpMessageInterface,
                                         MARTe2HTTPObjectBrowser)
from martepy.marte2.objects.referencecontainer import MARTe2ReferenceContainer
from martepy.marte2.objects.statemachine import MARTe2StateMachineEvent, MARTe2StateMachine
from martepy.marte2.objects.configuration_database import MARTe2ConfigurationDatabase
from martepy.marte2.objects.gam_scheduler import MARTe2GAMScheduler
from martepy.marte2.objects.real_time_state import MARTe2RealTimeState
from martepy.marte2.objects.real_time_thread import MARTe2RealTimeThread

__all__ = [
        'MARTe2Message',
        'MARTe2HttpDirectoryResource',
        'MARTe2HttpService',
        'MARTe2HttpMessageInterface',
        'MARTe2HTTPObjectBrowser',
        'MARTe2ReferenceContainer',
        'MARTe2StateMachineEvent','MARTe2StateMachine',
        'MARTe2ConfigurationDatabase',
        'MARTe2GAMScheduler',
        'MARTe2RealTimeState',
        'MARTe2RealTimeThread'
    ]



"""A more unified and simpler approach to developing MARTe2 application
"""
import copy
import os
import martepy.marte2.configwriting as marteconfig
from martepy.functions.gam_functions import (addAlias, addDimensions, assignUniqueName, consolidate,
                                             getAlias, getDatasource, getGamByName,
                                             getKeyAttribute, setDatasource,
                                             setKeyAttribute,
                                             removeKeysFromConfig)
from martepy.functions.extra_functions import getname
from martepy.marte2.gams.iogam import IOGAM
from martepy.marte2.datasources.gam_datasource import GAMDataSource
from martepy.marte2.factory import Factory
from martepy.marte2.type_database import TypeDBv2 as TypeDB
from martepy.marte2.datasources.files.file_datasources import RFileWriter
from martepy.marte2.datasources.files.writer import FileWriter
from martepy.marte2.qt_functions import generateUniqueName
from martepy.marte2.datasources.async_bridge import AsyncBridge

class MARTe2Exception(Exception):
    ''' A general MARTe2 Exception - should match similar to what you would receive
    if you tried to actually run the configuration in MARTe2.'''

class MARTe2Application():
    ''' The pythonic object representation of a MARTe2 application, how to build it and
    ultimately write this into it's string configuration format. '''
    def __init__(self, app_name = '$App'):
        self.resetApp()
        self.externals = self.NonDuplicatingList()
        self.inputs = []
        self.outputs = []
        self.additional_datasources = self.NonDuplicatingList()
        self.additional_logging = {'inputs':[],'outputs':[],'logger_in':[]}
        self.triggering_services = []
        self.functions = self.NonDuplicatingList()
        self.states = self.NonDuplicatingList()
        self.internals = self.NonDuplicatingList()
        self.objects = self.NonDuplicatingList()
        self.default_data_source = 'DDB0'
        self.config = {'log': True,
                       'log_level': 31,
                       'error_state': True,
                       'state_machine': True,
                       'numcores': 16,
                       'timefrequency': 1000}
        self.type_db = TypeDB()
        self.maxcycles = 40000
        self.app_name = 'App'
        # Load the factory
        self.factory = Factory()
        self.marte2_dir = os.path.abspath(os.path.join(os.path.dirname(__file__)))
        datasources = os.path.join(self.marte2_dir,"datasources", "datasources.json")
        gams = os.path.join(self.marte2_dir,"gams", "gams.json")
        objects = os.path.join(self.marte2_dir,"objects","objects.json")
        self.factory.loadRemote(datasources)
        self.factory.loadRemote(gams)
        self.app_name = app_name
        self.factory.loadRemote(objects)
        self.logging_iogam = IOGAM('+LoggingGAM',[],[])
        self.filewriter = RFileWriter('+LoggingFileWriter',filename='output.csv')
        self.asyncbridge = AsyncBridge('+LoggingAsyncBridge')
        self.async_to_io = IOGAM('+async_sim_io',[],[])
        self.io_asyncs = []
        self.loggingintoio = IOGAM('+ToLogGAM',[],[])
        self.iogams = [self.logging_iogam, self.loggingintoio, self.async_to_io]
        self._internal_datasources = []

    def loadTypeLibrary(self, filepath: str):
        ''' Load type definition directory into our type database '''
        self.type_db.loadDb(filepath)

    class NonDuplicatingList(list):
        ''' A non duplicating list class - effectively a set '''
        def append(self, item):
            if not any(item == list_item for list_item in self):
                super().append(item)

        def __iadd__(self, other):
            for item in other:
                self.append(item)
            return self

    def resetApp(self):
        ''' Reset our application instance. '''
        self.externals = self.NonDuplicatingList()
        self.inputs = []
        self.outputs = []
        self.additional_datasources = self.NonDuplicatingList()
        self.additional_logging = {'inputs':[],'outputs':[],'logger_in':[]}
        self.triggering_services = []
        self.functions = self.NonDuplicatingList()
        self.states = self.NonDuplicatingList()
        self.internals = self.NonDuplicatingList()
        self.objects = self.NonDuplicatingList()


    def add(self, externals=[], additional_datasources=[], functions=[],
            states=[], internals=[], objects=[]):
        ''' Safely add objects to our application '''
        self.externals += externals
        self.additional_datasources += additional_datasources
        self.functions += functions
        self.states += states
        self.internals += internals
        self.objects += objects

    def getSignals(self):
        """We want to produce a list of all signals made
        """
        signals = {}
        for anyproducer in self.additional_datasources + self.functions:
            for output_signal in anyproducer.output_signals:
                if 'DataSource' in output_signal[1]['MARTeConfig']:
                    datasource = output_signal[1]['MARTeConfig']['DataSource']
                else:
                    datasource = anyproducer.configuration_name.lstrip('+')
                if datasource not in signals:
                    signals[datasource] = {}
                signals[datasource][getAlias(output_signal)] = output_signal
        return signals

    def getInputDatasources(self):
        ''' Return a list of datasources which bring data into the application from the outside
        world '''
        # Need to try a different method, we need to collect what is our DDB GAM Datasources first
        # So we can ignore these and then any input signal from any other datasource can be assumed
        # to have come from the outside world, we collect these aliases, avoid duplicates
        gam_datasources = [gam_datasource for gam_datasource in self.additional_datasources if
                           isinstance(gam_datasource, GAMDataSource)]
        gam_names = [gam_datasource.configuration_name.strip('+') for gam_datasource
                     in gam_datasources]
        external_inputs = []
        for function in self.functions:
            for input_signal in function.input_signals:
                if input_signal[1]['MARTeConfig']['DataSource'] not in gam_names:
                    external_inputs.append(input_signal)

        return external_inputs

    def getOutputDatasources(self):
        ''' Return a list of datasources which output to the outside world '''
        # Need to try a different method, we need to collect what is our DDB GAM Datasources first
        # So we can ignore these and then any output signal to any other datasource can be assumed
        # to go to the outside world, we collect these aliases, avoid duplicates
        gam_datasources = [gam_datasource for gam_datasource in self.additional_datasources if
                           isinstance(gam_datasource, GAMDataSource)]
        gam_names = [gam_datasource.configuration_name.strip('+') for
                     gam_datasource in gam_datasources]
        external_outputs = []
        for function in self.functions:
            for output_signal in function.output_signals:
                if output_signal[1]['MARTeConfig']['DataSource'] not in gam_names:
                    external_outputs.append(output_signal)

        return external_outputs

    def getinputSignals(self):
        """We want to produce a list of all signals made
        """
        signals = {}

        for anyproducer in self.additional_datasources + self.functions:
            for output_signal in anyproducer.output_signals:
                signals[getAlias(output_signal)] = output_signal
        return signals

    def newUniqueName(self, basename):
        ''' Generate a unique name from what is already defined in our application. '''
        count = 0
        existing_signals = []
        name = basename
        for function in self.functions:
            for output_signal in function.output_signals:
                existing_signals.append(output_signal[0])
                existing_signals.append(getAlias(output_signal))
        if name in existing_signals:
            for _ in existing_signals: # It would be impossible to exceed this,
                # better than doing an infinite while loop
                count = count + 1
                name = f'{basename}{count}'
                if name not in existing_signals:
                    break
        return name

    def onlyErrors(self): # pylint:disable=R0914, R0915, R0912
        ''' Check our application for any errors given what we know, if we were to run this
        configuration now, what would MARTe2 error/say? '''
        exceptions = []
        consumed = {}
        produced = {}
        function_names = []
        # Signal names of same name within same function - use MARTeApplication.getAlias
        for function in self.functions:
            # Same named functions in application
            if function.configuration_name.lstrip('+') in function_names:
                exceptions.append(MARTe2Exception("""Two functions with the same name exist -
this can happen if you have two functions in the same state or across states which are
non-identical but identical in name."""))
            function_names.append(function.configuration_name.lstrip('+'))
            input_signal_names = []
            for signal in function.input_signals:
                if signal[0] in input_signal_names:
                    exceptions.append(MARTe2Exception(f"""Cannot have same root name of input
signal repeated within function: {function.configuration_name.lstrip('+')}"""))
                input_signal_names.append(signal[0])
                datasource = signal[1]['MARTeConfig']['DataSource']
                if datasource not in consumed:
                    consumed[datasource] = []
                consumed[datasource].append(getAlias(signal))
            output_signal_aliases = []
            output_signal_names = []
            for signal in function.output_signals:
                if getAlias(signal) in output_signal_aliases:
                    exceptions.append(MARTe2Exception(f"""Multiple producers of alias within
function: {function.configuration_name.lstrip('+')}"""))
                elif signal[0] in output_signal_names:
                    exceptions.append(MARTe2Exception(f"""Cannot have same root name of output
signal repeated within function: {function.configuration_name.lstrip('+')}"""))
                output_signal_aliases.append(getAlias(signal))
                output_signal_names.append(signal[0])
        # More than one producer of a signal in a datasource - use MARTeApplication.getAlias

        for datasource_name, datasource_signals in produced.items():
            seen = set()
            for string in datasource_signals:
                if string in seen:
                    exceptions.append(MARTe2Exception(f"""Multiple producers found of
signal/alias: {string} in datasource {datasource_name}"""))
                else:
                    seen.add(string)
        # Same named datasources in application
        datasource_names = []
        for datasource in self.additional_datasources:
            if datasource.configuration_name.lstrip('+') in datasource_names:
                exceptions.append(MARTe2Exception("""Two datasources with the same name exist
 - this can happen if you have two datasources in the same state or across states which are
 non-identical but identical in name."""))
            datasource_name = datasource.configuration_name.replace('+','')
            datasource_names.append(datasource_name)
            input_signal_names = []
            for signal in datasource.input_signals:
                if signal[0] in input_signal_names:
                    exceptions.append(MARTe2Exception(f"""Cannot have same root name of input
 signal repeated within datasource: {datasource.configuration_name.lstrip('+')}"""))
                input_signal_names.append(signal[0])

            output_signal_names = []
            for signal in datasource.output_signals:
                if signal[0] in output_signal_names:
                    exceptions.append(MARTe2Exception(f"""Cannot have same root name of
 output signal repeated within datasource: {datasource.configuration_name.lstrip('+')}"""))
                output_signal_names.append(signal[0])
            # A datasource must be read/written in each state
            for input_signal in datasource.input_signals:
                for state in self.states:
                    if not any(((sinput[1]['MARTeConfig']['DataSource'] ==
                                 datasource.configuration_name.lstrip('+') and
                                 getAlias(sinput) == getAlias(input_signal) for
                                 sinput in function.input_signals) for
                                 function in thread.functions) for
                                 thread in state.threads.objects):
                        exceptions.append(MARTe2Exception(f"""No function in state
{state.configuration_name.lstrip('+')} found reading from datasource
{datasource.configuration_name.lstrip('+')} for signal {getAlias(input_signal)}"""))
                    if datasource_name not in consumed:
                        consumed[datasource_name] = []
                    consumed[datasource_name].append(getAlias(input_signal))

            for output_signal in datasource.output_signals:
                for state in self.states:
                    if all(len(thread.functions) == 0 for thread in state.threads.objects):
                        continue
                    if not any(((output[1]['MARTeConfig']['DataSource'] ==
                                 datasource.configuration_name.lstrip('+') and
                                 getAlias(output) == getAlias(output_signal) for
                                 output in function.output_signals) for
                                 function in thread.functions) for
                                 thread in state.threads.objects):
                        exceptions.append(MARTe2Exception(f"""No function in state
{state.configuration_name.lstrip('+')} found writing to datasource
{datasource.configuration_name.lstrip('+')} for signal {getAlias(output_signal)}"""))
                if datasource_name not in produced:
                    produced[datasource_name] = []
                if getAlias(output_signal) not in produced[datasource_name]:
                    produced[datasource_name].append(getAlias(output_signal))
        # All consumers must have their producing signal - use MARTeApplication.getAlias
        for datasource_name, consumed_signals in consumed.items():
            for consumer_signal in consumed_signals:
                if datasource_name not in produced:
                    if datasource_name not in [getname(a) for a in self.additional_datasources]:
                        exceptions.append(MARTe2Exception(f"""No datasource {datasource_name}
 found yet consumed by signal {consumer_signal}"""))
                if datasource_name not in produced:
                    if datasource_name not in [getname(a) for a in self.additional_datasources]:
                        exceptions.append(MARTe2Exception(f"""datasource {datasource_name} not
  found for consumed signal {consumer_signal}"""))
                    continue
                if consumer_signal not in produced[datasource_name]:
                    exceptions.append(MARTe2Exception(f"""No producer found for signal
 {consumer_signal} in datasource {datasource_name}"""))

        # Now qualify that you can't have states with the same name and
        # subsequently threads with the same name
        state_names = []
        for state in self.states: # pylint: disable=R1702
            thread_names = []
            produced_state = copy.deepcopy(produced)
            if state.configuration_name.lstrip('+') in state_names:
                exceptions.append(MARTe2Exception(f"""Multiple states with the same
 name: {state.configuration_name.lstrip('+')}"""))
            state_names.append(state.configuration_name.lstrip('+'))
            for thread in state.threads.objects:
                if thread.configuration_name.lstrip('+') in thread_names:
                    exceptions.append(MARTe2Exception(f"""Multiple threads with the same
 name in state: {state.configuration_name.lstrip('+')},
 with name: {thread.configuration_name.lstrip('+')}"""))
                for function in thread.functions:
                    for output_signal in function.output_signals:
                        if 'DataSource' in signal[1]['MARTeConfig'].keys():
                            datasource = signal[1]['MARTeConfig']['DataSource']
                            if datasource not in produced_state:
                                produced_state[datasource] = []
                            produced_state[datasource].append(getAlias(signal))
                thread_names.append(thread.configuration_name.lstrip('+'))

        # Only one function can write to a FileWriter across states/threads
        for filewriter in [a for a in self.additional_datasources if
                           isinstance(a, (FileWriter, RFileWriter))]:
            writers = [a for a in self.functions if any(
                getKeyAttribute(output_signal, 'DataSource') ==
                getname(filewriter) for output_signal in a.output_signals)]
            if len(writers) > 1:
                exceptions.append(MARTe2Exception(f"""Multiple writers to FileWriter
 {getname(filewriter)}"""))

        return exceptions

    @staticmethod
    def _unflattenLog(orig_signal, type_db, loggingiogam, loggingintoio, filewriter, loggamsource): # pylint:disable=R0914
        # Here we want to get the signals associated type and
        # then unflatten it's members so we can get it into
        # the SimFileWriter
        new_gams = []
        types_used = []
        type_name = getKeyAttribute(orig_signal,'Type')
        type_def = type_db.getTypeByName(type_name)

        if type_def.name not in types_used:
            types_used += [type_def.name]

        top_gam = IOGAM(f'+unflatten{getAlias(orig_signal)}',
                        input_signals=[orig_signal], output_signals=[])
        new_gams = [top_gam]

        for field in type_def.fields:
            if not type_db.isFundamental(field.type):
                new_gams += MARTe2Application._unflattenLog(field, type_db, loggingiogam,
                                                            loggingintoio, filewriter,
                                                            loggamsource)
            else:
                # For each field, we now want to create an IOGAM output_signal for it
                alias = getAlias(orig_signal) + field.name
                out_sgl = (alias, copy.deepcopy(orig_signal[1]))
                setKeyAttribute(out_sgl,'Type',field.type)
                default = 0 if field.type in ('uint32','uint64', 'uint8') else 0.0
                setKeyAttribute(out_sgl,'Default',default)
                setKeyAttribute(out_sgl,'Alias',alias)
                top_gam.output_signals.append(out_sgl)
                loggingiogam.input_signals += [out_sgl]
                out_sgl = copy.deepcopy(out_sgl)
                if any(signal[0] == out_sgl[0] for signal in loggingiogam.output_signals):
                    existing_names = [a[0] for a in loggingiogam.input_signals]
                    out_sgl = (generateUniqueName(existing_names, out_sgl[0]), out_sgl[1])

                setKeyAttribute(out_sgl,'DataSource',loggamsource)

                loggingiogam.output_signals += [out_sgl]
                loggingintoio.input_signals += [out_sgl]
                out_sgl = copy.deepcopy(out_sgl)
                setKeyAttribute(out_sgl,'DataSource',getname(filewriter))

                loggingintoio.output_signals += [out_sgl]

                out_sgl = copy.deepcopy(out_sgl)
                removeKeysFromConfig(out_sgl, ['DataSource'])
                filewriter.input_signals += [out_sgl]
        return new_gams

    def _logSignal(self, signal, thread_index: int, state, thread): # pylint: disable=R0914,R0915
        ''' This function will log a signal, if it is outside the primary thread
        it will use an Async Bridge to transfer the signal to the primary and then 
        add this to the log. If it is a complex type, it will unflatten the signal type. '''
        # Is not a signal from the outside world and therefore we want to log it
        # into our primary thread and primary FileWriter for the simulation.
        new_functions = []
        osig_name = getAlias(signal)
        if thread_index == 0:
            # Now we need to add to loggingintoio IOGAM to get this into our
            # LoggingGAMSource

            if not any(getAlias(signal) == osig_name for
                        signal in self.logging_iogam.input_signals):

                new_signal = copy.deepcopy(signal)
                removeKeysFromConfig(new_signal, ['Default'])

                addDimensions(new_signal)

                addAlias(new_signal)

                out_signal = copy.deepcopy(new_signal)

                if self.type_db.isFundamental(
                    new_signal[1]['MARTeConfig']['Type']):

                    setDatasource(out_signal, 'LogGAMSource')

                    assignUniqueName(new_signal, self.logging_iogam.input_signals)

                    self.logging_iogam.input_signals += [new_signal]

                    assignUniqueName(out_signal, self.logging_iogam.output_signals)

                    removeKeysFromConfig(out_signal, ['Alias'])

                    self.logging_iogam.output_signals += [out_signal]

                    self.loggingintoio.input_signals += [out_signal]

                    out_signal = copy.deepcopy(out_signal)

                    setDatasource(out_signal, getname(self.filewriter))

                    self.loggingintoio.output_signals += [out_signal]

                    out_signal = copy.deepcopy(out_signal)

                    removeKeysFromConfig(out_signal, ['DataSource'])

                    self.filewriter.input_signals += [out_signal]
                else:
                    new_functions += MARTe2Application._unflattenLog(out_signal,
                                                                     self.type_db,
                                                                     self.logging_iogam,
                                                                     self.loggingintoio,
                                                                     self.filewriter,
                                                                     'LogGAMSource')
        else:
            # Create an IOGAM to get it to the bridge - if not already made
            # Now add signal to loggingintoio
            # Next check if not already logged
            # We need to signals, the input of our signal to the async bridge from the thread
            # and the output signal to the async bridge
            input_signal = copy.deepcopy(signal)
            # Neither signal should have a Default value set in the IOGAM
            removeKeysFromConfig(input_signal, ['Default'])

            # We assume here that the AsyncBridge can handle complex types
            # so we will unflatten this later in the primary thread, set the Async Bridge
            state_name = getname(state)
            thread_name = getname(thread)
            iobridge_name = f'+{state_name}_{thread_name}_async_sim'
            # Get the Async Bridge in question
            iobridge = getGamByName(iobridge_name.lstrip('+'), self.io_asyncs)
            # Ensure it has a unique name
            if iobridge:
                assignUniqueName(input_signal, iobridge.input_signals)
            # Make sure it will still link by alias back to the original
            input_alias = getAlias(signal)
            input_signal[1]['MARTeConfig']['Alias'] = input_alias
            # Set the output signal to the async bridge
            out_signal = copy.deepcopy(input_signal)
            # Now create our output_signal and set it to the bridge
            setDatasource(out_signal, 'LoggingAsyncBridge')
            removeKeysFromConfig(out_signal, ['Alias'])
            # Now we're ready, add to our IOGAM to Async Bridge
            io_async_gam = consolidate(self.io_asyncs,'IOGAM',iobridge_name,
                                        [input_signal],[out_signal])
            # Make sure we keep track of our async IO's
            if io_async_gam not in self.io_asyncs:
                self.io_asyncs.append(io_async_gam)
            if io_async_gam not in new_functions:
                new_functions.append(io_async_gam)
            # Now we know what the signal will be on the primary thread of the async bridge
            # Now create the logging of this signal in the primary thread
            tolog = copy.deepcopy(out_signal)
            assignUniqueName(tolog, self.loggingintoio.input_signals)
            setDatasource(tolog, 'LogGAMSource')
            consolidate(state.threads.objects[0].functions,
                        'IOGAM','+async_sim_io',
                        [out_signal],[tolog])

            if self.type_db.isFundamental(input_signal[1]['MARTeConfig']['Type']):
                # Add to our log

                self.loggingintoio.input_signals += [tolog]

                tolog = copy.deepcopy(tolog)
                setDatasource(tolog, getname(self.filewriter))

                self.loggingintoio.output_signals += [tolog]
                tolog = copy.deepcopy(tolog)
                del tolog[1]['MARTeConfig']['DataSource']
                self.filewriter.input_signals += [tolog]
            else:
                # Unflatten and add to log
                MARTe2Application._unflattenLog(tolog,
                                                self.type_db,
                                                self.logging_iogam,
                                                self.loggingintoio,
                                                self.filewriter,
                                                'LogGAMSource')

        return new_functions

    def buildLog(self):
        """Build a datalogger IO and log signals from our application given user input
        """
        # Now that everything exists as it should, we now need to create a
        # method of logging every signal ever created.
        # First we need to track every signal that is produced, if it is not
        # a part of our constant or misc DDB's
        # We should start by creating a IOGAM called loggingGAM which feeds
        # from our logDDB into our our FileWriter
        # We should also create our FileWriter as a reference, only one IOGAM
        # can write to this FileWriter, hence our approach
        #
        # When we move signals into our logDDB we should check if it already
        # exists as a signal name duplicate, if so, generate unique name

        self.additional_datasources += [self.asyncbridge, self.filewriter]
        self.functions += [self.logging_iogam, self.loggingintoio]
        for state in self.states:
            if len(state.threads.objects) > 0:
                state.threads.objects[0].functions += self.iogams
            for thread_index, thread in enumerate(state.threads.objects):
                new_functions = []
                # if we are not in thread 0 we must create an AsyncBridge for this
                # thread if not already existing and send this signal through said bridge
                # Once we have all these signals into thread 0, we use an IOGAM for each
                # state to transfer these into our log DDB and then add these to our
                # loggingIOGAM and filewriter

                for function in thread.functions:
                    if function in self.iogams or function in self.io_asyncs:
                        continue
                    if hasattr(function, 'isSimulation') and function.isSimulation:
                        continue
                    for output_signal in function.output_signals:
                        if getDatasource(output_signal) not in self._internal_datasources:
                            new_functions += self._logSignal(output_signal,
                                                              thread_index, state, thread)
                thread.functions += new_functions

    def writeToConfig(self):
        ''' Write our application out and return as a string '''
        # Fix any breaks in the config automatically
        self.sanitize()
        self.removeUnused()

        out = marteconfig.StringConfigWriter()

        for piece in self.externals:
            if piece:
                piece.write(out)

        #Adding capability to have non application and non state machine
        # entities in our configuration, for instance, the HTTPObjectBrowser
        for obj in self.objects:
            obj.write(out)

        out.startClass(f'${self.app_name.replace("$","")}', 'RealTimeApplication')

        out.startClass('+Functions', 'ReferenceContainer')

        for i in self.functions:
            i.write(out)

        out.endSection('+Functions')

        out.startClass('+Data', 'ReferenceContainer')

        if not len([a for a in self.additional_datasources if isinstance(a, GAMDataSource)]) == 0:
            out.writeNode('DefaultDataSource', self.default_data_source)

        for i in self.additional_datasources:
            i.write(out)

        out.endSection('+Data')

        out.startClass('+States', 'ReferenceContainer')

        for i in self.states:
            i.write(out)

        out.endSection('+States')

        for i in self.internals:
            i.write(out)

        out.endSection(f'${self.app_name.replace("$","")}')

        return str(out) + "\n"

    def sanitize(self):
        ''' Resanitize an application between its state, 
        threads functions and it's functions definitions.'''
        self.functions = []
        for state in self.states:
            for thread in state.threads.objects:
                for function in thread.functions:
                    if function not in self.functions: # Avoid duplications
                        self.functions.append(function)


    def removeUnused(self):
        ''' Removes empty and unused objects such as:
        - An IOGAM with no inputs or outputs
        - An Async bridge with no IOGAMs writing to it or from it
        - A GAM DataSource with no signals defined throughout the application into it. '''
        self.sanitize()
        # If a GAM Datasource has no values assigned to it - we should remove it
        gam_sources = [(a.configuration_name.lstrip('+'), a) for
                       a in self.additional_datasources if isinstance(a, GAMDataSource)]
        for datasource, gam_object in gam_sources:
            if not any(any(output_signal[1]['MARTeConfig']['DataSource'] == datasource for
                           output_signal in function.output_signals) for
                           function in self.functions):
                # We need to remove this
                self.additional_datasources.remove(gam_object)

        # If an IOGAM has no inputs/outputs, it should be removed,
        # if it has no inputs, sanity is all but lost:
        for function in self.functions:
            if isinstance(function, IOGAM):
                if len(function.input_signals) == 0 or len(function.output_signals) == 0:
                    # Delete the function from both our functions and
                    # whereever it's used in threads/states
                    for state in self.states:
                        for thread in state.threads.objects:
                            thread.functions = [a for a in thread.functions if not a == function]

        self.sanitize()

        # As above, if an async bridge exists without any
        # reading/writing signals the heap allocation will fail
        asyncbridges = [a for a in self.additional_datasources if isinstance(a, AsyncBridge)]
        for bridge in asyncbridges:
            if not any(any(input_signal[1]['MARTeConfig']['DataSource'] == getname(bridge) for
                           input_signal in function.input_signals) for
                           function in self.functions) and not any(
                any(output_signal[1]['MARTeConfig']['DataSource'] == getname(bridge) for
                    output_signal in function.output_signals) for function in self.functions):

                # Remove the bridge as it is unused
                self.additional_datasources.remove(bridge)



''' Pythonic representation of the Event Condition Trigger Object '''

from martepy.marte2.config_object import MARTe2ConfigObject
from martepy.functions.extra_functions import getname

class MARTe2EventConditionTrigger(MARTe2ConfigObject):
    ''' Pythonic representation of the Event Condition Trigger Object '''
    def __init__(self,
                    configuration_name: str = '+NewEvent',
                    class_name = "EventConditionTrigger",
                    eventtriggers = {},
                    msgs = [],
                ):
        super().__init__()
        self.class_name = class_name
        self.eventtriggers = eventtriggers
        self.configuration_name = configuration_name
        self.objects = msgs

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            attributes = ['configuration_name', 'eventtriggers', 'objects']
            for item in attributes:
                if not getattr(self, item) == getattr(other, item):
                    return False
            return True
        return False

    def write(self, config_writer):
        ''' Write the configuration of our object '''
        config_writer.startClass('+' + getname(self), self.class_name)
        config_writer.startSection('EventTriggers')
        for key, value in self.eventtriggers.items():
            config_writer.writeNode(key, str(value))
        config_writer.endSection('EventTriggers')
        for obj in self.objects:
            obj.write(config_writer)
        config_writer.endSection('+' + getname(self))

    def serialize(self):
        ''' Serialize the object '''
        res = super().serialize()
        res['class_name'] = "EventConditionTrigger"
        res['eventtriggers'] = self.eventtriggers
        res["objects"] = [a.serialize() for a in self.objects]
        return res

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True, # pylint: disable=W0221
                    factory=None) -> bool:
        ''' Deserialize the given data to our class instance '''
        def createObj(a):
            return factory.create(a['class_name'])().deserialize(a, factory=factory)
        super().deserialize(data, hashmap, restore_id, factory=factory)
        self.class_name = "EventConditionTrigger"
        self.eventtriggers = dict(data['eventtriggers'])
        # This part is tricky as a ref container can
        # contain any object so the factory needs to recircle here
        self.objects = [createObj(a) for a in data["objects"]]
        return self

def initialize(factory, plugin_datastore) -> None:
    ''' Register the object with the factory '''
    factory.registerBlock("EventConditionTrigger", MARTe2EventConditionTrigger, plugin_datastore)



''' Import most of the useful datasources in one go '''

from martepy.marte2.datasources.gam_datasource import GAMDataSource
from martepy.marte2.datasources.logger_datasource import LoggerDataSource
from martepy.marte2.datasources.timing_datasource import TimingDataSource
from martepy.marte2.datasources.udp import UDPReceiver, UDPSender
from martepy.marte2.datasources.linux_timer import LinuxTimer
from martepy.marte2.datasources.files import FileReader, FileWriter
from martepy.marte2.datasources.sdn import SDNPublisher, SDNSubscriber
from martepy.marte2.datasources.epics import EPICSPublisher, EPICSSubscriber
from martepy.marte2.datasources.async_bridge import AsyncBridge
from martepy.marte2.datasources.files.file_datasources import RFileReader, RFileWriter

__all__ = [
        'GAMDataSource',
        'LoggerDataSource',
        'TimingDataSource',
        'udp',
        'LinuxTimer',
        'files',
        'sdn',
        'epics',
        'AsyncBridge',
        'RFileReader',
        'RFileWriter'
    ]



"""gam_functions.py contains functions that create or maniuplate gams"""

# Is necessary to import all for consolidate function, don't know what GAM the user will
# consolidate
from martepy.marte2.gams import * # pylint: disable=W0614
from martepy.marte2.qt_functions import generateUniqueName

def assignUniqueName(signal: tuple, signals_list: list):
    ''' Ensures that a signal is uniquely named compared to a list of signals '''
    if any(signal_item[0] == signal[0] for signal_item in signals_list):
        existing_names = [a[0] for a in signals_list]
        alias_name = getAlias(signal)
        signal[1]['MARTeConfig']['Alias'] = alias_name
        signal = (generateUniqueName(existing_names, signal[0]), signal[1])
    return signal

def removeKeysFromConfig(signal: tuple, keys: list):
    ''' Safely Delete keys from a signal '''
    for key in keys:
        if key in signal[1]['MARTeConfig'].keys():
            del signal[1]['MARTeConfig'][key]

def getKeyAttribute(signal, key):
    ''' Given a specific MARTe2Config key, retrieve the value '''
    return signal[1]['MARTeConfig'][key]

def setKeyAttribute(signal, key, value):
    ''' Set a attribute of the MARTe Config'''
    signal[1]['MARTeConfig'][key] = value

def addDimensions(signal: tuple):
    ''' If a signal does not contain elements or dimensions, add them '''
    if 'NumberOfDimensions' not in signal[1]['MARTeConfig'].keys():
        signal[1]['MARTeConfig']['NumberOfDimensions'] = '1'

    if 'NumberOfElements' not in signal[1]['MARTeConfig'].keys():
        signal[1]['MARTeConfig']['NumberOfElements'] = '1'

def addAlias(signal: tuple):
    ''' If an Alias is not set, set it to the basic name '''
    if 'Alias' not in signal[1]['MARTeConfig'].keys():
        alias_name = getAlias(signal)
        signal[1]['MARTeConfig']['Alias'] = alias_name

def getGamByName(name: str, function_list):
    ''' Returns a GAM based on the name in a given function list '''
    def getname(value):
        return value.configuration_name.lstrip('+')
    return next((a for a in function_list if getname(a) == name), None)

def getAlias(signal_key):
    ''' Return the alias of a signal '''
    if 'Alias' in list(signal_key[1]['MARTeConfig'].keys()):
        return signal_key[1]['MARTeConfig']['Alias']
    return signal_key[0]

def setDatasource(signal, datasource):
    ''' Simplification of setting a signals datasource '''
    signal[1]['MARTeConfig']['DataSource'] = datasource

def getDatasource(signal):
    ''' Simplification of getting a signals datasource '''
    return signal[1]['MARTeConfig']['DataSource']


def isMatchingSignal(signal, other_signal):
    ''' Here we say a signal is the same if it has the same alias and 
    datasource where the signal exists.'''
    if not getAlias(signal) == getAlias(other_signal):
        return False
    if not getDatasource(signal) == getDatasource(other_signal):
        return False
    return True

def findSignalInFunctionList(signal, functions):
    ''' Will return the signal that matches a given signal in a list of functions '''
    try:
        return next(
            producer
            for function in functions
            for producer in function.output_signals
            if isMatchingSignal(producer, signal)
        )
    except StopIteration:
        return None

def getParameterName(parameter_name: str):
    """This takes the serialized name of a parameter and returns it's attribute name"""
    return parameter_name.lower().replace(' ','_')

def consolidate(thread_0_functions, gam_type, config_name, inputs=[], outputs=[],prepend=False): # pylint: disable=R0912
    """:param thread_0_functions: list of references to rtcc1blocks specific class
    :type thread_0_functions: list[]
    :param str gam_type: type of gam
    :param str config_name: SDN config
    :param inputs: input signal
    :type inputs: list[tuple[str, dict[]]]
    :param outputs: output signal
    :type outputs: list[tuple[str, dict[]]]

    :return gam: marte2 gam
    :rtype: marte2 gam

    Add a GAM which creates a constant signal or reads a signal from SDN
    but check first if a similarly named GAM of that type exists, and if it
    does simply add the signals provided to this method into the
    pre-existing GAM.
    """
    if not any(
        a.class_name == gam_type and a.configuration_name == config_name
        for a in thread_0_functions
    ):
        # First actual IOGAM request for any SDN signal so let's add it.
        gam = globals()[gam_type]
        if gam_type == "ConstantGAM":
            if prepend:
                thread_0_functions.insert(0,gam(configuration_name=config_name, output_signals=[]))
            else:
                thread_0_functions.append(gam(configuration_name=config_name, output_signals=[]))
        elif prepend:
            thread_0_functions.insert(0,
                gam(configuration_name=config_name, input_signals=[], output_signals=[])
            )
        else:
            thread_0_functions.append(
                gam(configuration_name=config_name, input_signals=[], output_signals=[])
            )
    index = next(
        i for i, d in enumerate(thread_0_functions) if d.configuration_name == config_name
    )
    gam = thread_0_functions[index]
    if gam_type != "ConstantGAM": # pylint: disable=R1702
        if gam_type != "IOGAM":
            for inputsig, output in zip(inputs, outputs):
                # Here we test that if an output signal is unique, but the input signal is not
                # We still add the signal but give the input a unique signalname as well.
                # This is useful when we're assuming here that the output signal is a converted
                # signal type and the user would like to convert the same signal into multiple
                # formats this simplifies that process for them by maintaining the alias whilst
                # giving it a different signal name.
                if not any(a == output for a in gam.output_signals):
                    # Input signal does not yet exist, we can also presume output doesn't either.
                    # Let's add it - then reinsert our converting GAM to the index, replacing it.
                    # Check that our input signal name is unique and if not, make it unique
                    if any(a[0] == inputsig[0] for a in gam.input_signals):
                        i = 1
                        while(any(inputsig[0] + "_" + str(i) == a[0] for a in gam.input_signals)):
                            i = i + 1
                        inputsig = (inputsig[0] + "_" + str(i),inputsig[1])
                    gam.input_signals.append(inputsig)
                    gam.output_signals.append(output)
        else:
            for a in inputs:
                gam.input_signals.append(a)
            for a in outputs:
                gam.output_signals.append(a)
    else:
        for output in outputs:
            if not any(a == output for a in gam.output_signals):
                # Input signal does not yet exist, we can also presume output doesn't either.
                # Let's add it - then reinsert our convertin GAM to the index, replacing it.
                gam.output_signals.append(output)
    thread_0_functions[index] = gam
    return gam



''' Topologically sorts the dependencies of headers in a given directory based on header 
imports '''

import os
import re

def _extractIncludes(file_path):
    ''' Get the includes from a C++ header '''
    with open(file_path, 'r', encoding='utf-8') as file:
        content = file.read()
        # Use a regular expression to extract included header files
        includes = re.findall(r'#include\s+[<"]([^>"]+)[>"]', content)
        return includes

def _findDependencies(directory):
    ''' From a directory, create a dictionary of headers and their dependencies '''
    # Get all header files in the directory
    header_files = [file for file in os.listdir(directory) if file.endswith(".h")]

    # Create a dictionary to store dependencies
    dependencies = {file: set(_extractIncludes(os.path.join(directory, file))) for
                    file in header_files}

    return dependencies

def _topologicalSort(dependencies, directory):
    ''' Sort our dictionary based on it's dependencies defined from includes '''
    sorted_files = []
    visited = set()

    def visit(file):
        if file not in visited:
            visited.add(file)
            if os.path.exists(os.path.join(directory, file)):
                sorted_files.append(os.path.join(directory, file))

    for file in dependencies.keys():
        visit(file)

    return sorted_files[::-1]

def sortHeaders(directory):
    ''' Publically accessible function to sort headers and return an ordered list of 
    files given a directory '''
    dependencies = _findDependencies(directory)
    ordered_files = _topologicalSort(dependencies, directory)
    ordered_files.reverse()
    return ordered_files



''' Pythonic representation of ReferenceContainer '''
from martepy.marte2.config_object import MARTe2ConfigObject
from martepy.functions.extra_functions import getname

class MARTe2ReferenceContainer(MARTe2ConfigObject):
    ''' Pythonic representation of ReferenceContainer '''
    def __init__(self,
                    configuration_name: str = 'Message',
                    objects = [],
                ):
        super().__init__(
            )
        self.class_name = "ReferenceContainer"
        self.objects = objects
        self.configuration_name = configuration_name

    def write(self, config_writer):
        ''' Write our configuration of this class '''
        config_writer.startClass('+' + getname(self),"ReferenceContainer")
        for obj in self.objects:
            obj.write(config_writer)
        config_writer.endSection('+' + getname(self))

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            if self.configuration_name == other.configuration_name:
                if self.objects == other.objects:
                    return True
        return False

    def serialize(self):
        ''' Serialize our object '''
        res = super().serialize()
        res['class_name'] = self.class_name
        res["objects"] = [a.serialize() for a in self.objects]
        return res

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True, #pylint: disable=W0221
                    factory=None) -> bool:
        ''' Deserialize our data into an instance of our class '''

        def createObj(a):
            return factory.create(a['class_name'])().deserialize(a, factory=factory)
        super().deserialize(data, hashmap, restore_id, factory=factory)
        self.class_name = data["class_name"]
        # This part is tricky as a ref container can
        # contain any object so the factory needs to recircle here
        self.objects = [createObj(a) for a in data["objects"]]
        return self

def initialize(factory, plugin_datastore) -> None:
    ''' Register us with the factory '''
    factory.registerBlock("ReferenceContainer", MARTe2ReferenceContainer, plugin_datastore)



''' Pythonic representation of the Filter GAM'''

from martepy.marte2.gam import MARTe2GAM
from martepy.marte2.qt_functions import (addInputSignalsSection,
                                         addOutputSignalsSection,
                                         addLineEdit,
                                         addComboEdit)

class FilterGAM(MARTe2GAM):
    ''' Pythonic representation of the Filter GAM'''
    def __init__(self,
                    configuration_name: str = 'Filter',
                    input_signals: list = [],
                    output_signals: list = [],
                    num = [],
                    den = [],
                    resetineachstate = 0,
                ):
        super().__init__(
                configuration_name = configuration_name,
                class_name = 'FilterGAM',
                input_signals = input_signals,
                output_signals = output_signals,
            )
        self.num = num
        self.den = den
        self.resetineachstate = resetineachstate

    def writeGamConfig(self, config_writer):
        ''' Write the Filter GAM Configuration '''
        reset_state = '0' if str(self.resetineachstate) == '0' else '1'
        config_writer.writeNode("Num",'{' + self.num.replace('{','').replace('}','') + '}')
        config_writer.writeNode("Den",'{' + self.den.replace('{','').replace('}','') + '}')
        config_writer.writeNode("ResetInEachState", reset_state)

    def serialize(self):
        ''' Serialize the object '''
        res = super().serialize()
        res['parameters']['num'] = self.num
        res['parameters']['den'] = self.den
        res['parameters']['resetineachstate'] = self.resetineachstate
        return res

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True) -> bool:
        ''' Deserialize the given object to our class instance '''
        super().deserialize(data, hashmap, restore_id)
        # Now we build up
        self.resetineachstate = data['parameters']["resetineachstate"]
        self.num = data['parameters']["num"]
        self.den = data['parameters']["den"]
        return self

    @staticmethod
    def loadParameters(mainpanel_instance, node):
        """This function is intended to be for the GUI where it can
        call the static instance of the class directly to generate
        the appropriate parameter modifier for the node in XMARTe2.
        """
        app_def = mainpanel_instance.parent.API.getServiceByName('ApplicationDefinition')
        datasource = app_def.configuration['misc']['gamsources'][0]
        addInputSignalsSection(mainpanel_instance, node, False)

        addOutputSignalsSection(mainpanel_instance, node, 3, False, datasource=datasource)

        addLineEdit(mainpanel_instance, node, "Amplitude: ", 'num', 3, 0)

        addLineEdit(mainpanel_instance, node, "Frequency: ", 'den', 3, 2)

        addComboEdit(mainpanel_instance, node, "Reset in Each State: ",
                     'resetineachstate', 4, 0, ['1','0'])

def initialize(factory, plugin_datastore) -> None:
    ''' Initialize the object with the factory '''
    factory.registerBlock("FilterGAM", FilterGAM, plugin_datastore)



''' Pythonic representation of the state machine object '''

from martepy.marte2.config_object import MARTe2ConfigObject
from martepy.functions.extra_functions import getname

class MARTe2StateMachine(MARTe2ConfigObject):
    ''' Pythonic representation of the state machine object '''
    def __init__(self,
                    configuration_name: str = 'GOTOSTATE1',
                    states = [],
                ):
        super().__init__(
            )
        self.class_name = "StateMachine"
        self.configuration_name = configuration_name
        self.states = states

    def addstate(self,state: list):
        ''' Safe method to add states to the machine '''
        self.states += state

    def write(self, config_writer):
        ''' Write the configuration of our object '''
        config_writer.startClass('+' + getname(self),"StateMachine")
        for state in self.states:
            state.write(config_writer)
        config_writer.endSection('+' + getname(self))

    def serialize(self):
        ''' Serialize the object '''
        res = super().serialize()
        res['class_name'] = self.class_name
        res["states"] = [a.serialize() for a in self.states]
        return res

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True, # pylint: disable=W0221
                    factory=None) -> bool:
        ''' Deserialize the given data to our class instance '''
        def createObj(a):
            return factory.create(a['class_name'])().deserialize(a, factory=factory)
        super().deserialize(data, hashmap, restore_id, factory=factory)
        self.class_name = data['class_name']
        self.states = [createObj(a) for a in data['states']]
        return self

def initialize(factory, plugin_datastore) -> None:
    ''' Register the object with the factory '''
    factory.registerBlock("StateMachine", MARTe2StateMachine, plugin_datastore)



'''
This module is a collection of classes used in the MARTe2 Python repository 
and GUI associated with MARTe2.
'''
import copy

from PyQt5.QtCore import QEvent, QTimer, Qt, pyqtSignal
from PyQt5.QtWidgets import (QCompleter,
                             QHBoxLayout,
                             QLineEdit,
                             QListWidget,
                             QPushButton,
                             QScrollArea,
                             QSizePolicy,
                             QSplitter,
                             QVBoxLayout,
                             QWidget)

from PyQt5.QtWidgets import (
    QGridLayout, QGroupBox, QMessageBox, QMainWindow
)
from martepy.functions.extra_functions import getname
from martepy.marte2.objects import MARTe2Message, MARTe2ConfigurationDatabase
from martepy.marte2.qt_functions import (createGridComboEdit,
                                         createGridLineEdit,
                                         generateUniqueName)

class AddRemoveHBtn(QWidget):
    ''' A generic QWidget for Add, Remove Buttons'''
    def __init__(self, parent=None):
        # We expose the add and remove btn's for the user to then create the connected
        # operations
        super().__init__(parent)
        self.hlayout = QHBoxLayout()
        self.setLayout(self.hlayout)
        self.add_btn = QPushButton("Add")
        self.remove_btn = QPushButton("Remove")
        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum)
        self.hlayout.addWidget(spacer)
        self.hlayout.addWidget(self.add_btn)
        self.hlayout.addWidget(self.remove_btn)

# Define a custom event type
class CustomEvent(QEvent): # pylint: disable=R0903
    ''' A Custom Event override '''
    def __init__(self, event_type):
        super().__init__(QEvent.User)
        self.event_type = event_type

class AutoCompleteLineEdit(QLineEdit):
    ''' This is an autocompleting line edittor for use in the TypeDB '''
    mouse_clicked = pyqtSignal()
    def __init__(self, window, items, app):
        super().__init__()
        self.app = app
        self.window = window
        self.items = items
        self.previous_text = ''
        self.initUI()

    def initUI(self):
        ''' Initialise our UI components '''
        self.completer = QCompleter(self.items)
        self.setCompleter(self.completer)
        self.completer.setCaseSensitivity(Qt.CaseInsensitive)
        self.completer.setCompletionMode(QCompleter.UnfilteredPopupCompletion)

        self.previous_text = ""
        self.mouse_clicked.connect(self.validateInput)
        # Connect textChanged signal to validation slot
        self.textChanged.connect(self.validateInput)

    def mousePressEvent(self, event):
        ''' User is inputting data '''
        if event.button() == Qt.LeftButton:
            # Handle the left mouse button click event
            # Create a custom event and post it to the application
            custom_event = CustomEvent("CustomEventType")
            self.app.postEvent(self.window, custom_event)
        super().mousePressEvent(event)

    # Override the customEvent method to handle custom events
    def customEvent(self, event):
        ''' Override our custom event handler to handle the user input and autocomplete '''
        if event.type() == QEvent.User:
            if event.event_type == "CustomEventType":
                self.completer.complete()

    def validateInput(self):
        ''' Check our input is correct and autocomplete if so. '''
        text = self.text()
        if not text:
            self.completer.complete()
            return

        valid = False
        for item in self.items:
            if text in item:
                valid = True
                break

        if not valid:
            self.setText(self.previous_text)
        else:
            self.previous_text = text

class PanelledListConfig(QWidget):
    ''' This is a generic class widget which splits a window into two sections and 
    create a list object on the left hand panel, leaving the right hand panel open for the user '''
    def __init__(self, parent=None, left_proportion=0.2, right_proportion=0.8, make_list=True):
        super().__init__(parent)
        # VLayout is the top level layout, if you use
        # this in a window and want say, save and close buttons underneath, add it to vlayout.
        self.v_layout = QVBoxLayout()
        self.setLayout(self.v_layout)
        self.splitter = QSplitter()
        self.v_layout.addWidget(self.splitter)
        # Now we have our splitter we need to create our two
        # widgets which will store the layouts for our left and right side panels

        self.left_panel = QWidget()
        self.left_panel_vlayout = QVBoxLayout()
        self.left_panel.setLayout(self.left_panel_vlayout)
        if make_list:
            self.left_list = QListWidget()
            self.left_list.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Expanding)
            self.left_panel_vlayout.addWidget(self.left_list)
            self.left_list.itemSelectionChanged.connect(self.scheduleEnsureSelection)
        self.splitter.addWidget(self.left_panel)
        # Right panel
        # We leave this blank with a VBoxLayout, the user should primarily interact with
        # the layout however can use the widget to replace it
        self.scroll_area = QScrollArea()
        self.right_panel_wgt = QWidget()
        self.right_panel_vlayout = QVBoxLayout()
        self.right_panel_wgt.setLayout(self.right_panel_vlayout)
        self.right_panel_wgt.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.scroll_area.setWidgetResizable(True)
        self.scroll_area.setWidget(self.right_panel_wgt)
        self.scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.scroll_area.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.splitter.addWidget(self.scroll_area)
        self.splitter.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        # Now set user defined proportions
        self.setProportions(left_proportion, right_proportion)

    def scheduleEnsureSelection(self):
        ''' Left list item has changed - most users override this method. '''
        # Use QTimer.singleShot to defer the ensure_selection method
        QTimer.singleShot(0, self.ensureSelection)

    def ensureSelection(self):
        ''' Make sure an actual item is selected and if not, select the first item. '''
        if not self.left_list.selectedItems():
            # If no items are selected, select the first item
            if self.left_list.count() > 0:
                self.left_list.setCurrentRow(0)

    def setProportions(self, left_proportion=0.2, right_proportion=0.8):
        ''' Setup the proportions of our splitter '''
        total_width = self.splitter.width()
        size1 = int(left_proportion * total_width)
        size2 = int(right_proportion * total_width)
        self.splitter.setSizes([size1,size2])

class MsgInfoWidget(QWidget):
    ''' Simplification of the MsgConfigWindow, QWidget which holds the Msg Info. '''
    def __init__(self, parent, msg_namechg, dest_chg, func_chg, mode_chg, wait_chg):
        super().__init__(parent)
        grid_layout = QGridLayout()
        self.setLayout(grid_layout)

        self.msg_name = createGridLineEdit(grid_layout, 0, 'Message Name:', msg_namechg,
                                            finished=True,space=True)
        self.msg_dest = createGridLineEdit(grid_layout, 1, 'Destination:', dest_chg,
                                            space=True)
        self.func = createGridLineEdit(grid_layout, 2, 'Function:', func_chg, space=True)
        self.mode = createGridComboEdit(grid_layout, 3, 'Mode:', mode_chg,
                                        ['ExpectsReply', 'ExpectsIndirectReply'], space=True)
        self.maxwait = createGridLineEdit(grid_layout, 4, 'MaxWait:', wait_chg,
                                            finished=True, space=True)

class MessageConfigWindow(QMainWindow):
    ''' The General Message Config Window which is used in multiple places
    to configure messages in the GUI '''
    def __init__(self, application, app, original, origin, title, close_inst=None,# pylint: disable=R0914, R0915
                 conversion_func=None, sizes=[0.25,0.25,0.5,0.5]) -> None:
        super().__init__()
        self.application = application
        self.app = app
        self.messages = []
        self.setWindowModality(Qt.ApplicationModal)
        self.close_inst = close_inst
        self.conversion_func = conversion_func
        # Set Window Size
        self.setSize(sizes[0],sizes[1],sizes[2],sizes[3])
        self.original = original
        self.origin = origin

        self.setWindowTitle(title)
        self.main_wgt = PanelledListConfig(self, 0.25, 0.75)
        self.central_layout = self.main_wgt.v_layout
        self.setCentralWidget(self.main_wgt)

        self.add_remmsg = AddRemoveHBtn(self.main_wgt.left_panel)
        self.main_wgt.left_panel_vlayout.addWidget(self.add_remmsg)
        self.add_remmsg.add_btn.clicked.connect(self.addMsg)
        self.add_remmsg.remove_btn.clicked.connect(self.remMsg)
        msg_box = QGroupBox("Message")
        # Create a layout for the QGroupBox
        msg_box_layout = QVBoxLayout(msg_box)
        # Set the layout of the QGroupBox
        msg_box.setLayout(msg_box_layout)
        msg_box.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.main_wgt.right_panel_vlayout.addWidget(msg_box)

        # Populate msg_box section
        self.grid_wgt = MsgInfoWidget(self, self.msgNameChg, self.destChg, self.funcChg,
                                      self.modeChg, self.waitChg)

        msg_box_layout.addWidget(self.grid_wgt)

        params_box = QGroupBox("Parameters")
        params_box_layout = QVBoxLayout(msg_box)

        params_box.setLayout(params_box_layout)

        self.param_config = PanelledListConfig(params_box,0.25,0.75)
        params_box_layout.addWidget(self.param_config)
        self.param_btns = AddRemoveHBtn()
        self.param_config.left_panel_vlayout.addWidget(self.param_btns)
        self.param_btns.add_btn.clicked.connect(self.addParam)
        self.param_btns.remove_btn.clicked.connect(self.remParam)
        # Right section
        grid_wgt = QWidget()
        grid_layout = QGridLayout()
        grid_wgt.setLayout(grid_layout)

        self.paramname = createGridLineEdit(grid_layout, 0, 'Parameter Name:',
                                               self.paramNameChg, finished=True, space=True)
        self.paramval = createGridLineEdit(grid_layout, 1, 'Parameter Value:',
                                              self.paramValChg, space=True)
        vspacer = QWidget()
        vspacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.param_config.right_panel_vlayout.addWidget(grid_wgt)
        self.param_config.right_panel_vlayout.addWidget(vspacer)

        msg_box_layout.addWidget(params_box)

        vspacer = QWidget()
        vspacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        msg_box_layout.addWidget(vspacer)

        btn_holder = QWidget()
        btn_hlayout = QHBoxLayout()
        btn_holder.setLayout(btn_hlayout)
        self.save_btn = QPushButton("Save")
        self.cancel_btn = QPushButton("Cancel")
        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum)
        btn_hlayout.addWidget(spacer)
        self.save_btn.clicked.connect(self.save)
        self.cancel_btn.clicked.connect(self.close)
        btn_hlayout.addWidget(self.save_btn)
        btn_hlayout.addWidget(self.cancel_btn)

        self.main_wgt.right_panel_vlayout.addWidget(btn_holder)

        self.param_config.left_list.currentItemChanged.connect(self.paramChanged)
        self.main_wgt.left_list.currentItemChanged.connect(self.msgChanged)
        self.loadMessages(copy.deepcopy(self.original))
        self.ignore_close_event = False
        self.show()

    def save(self):
        ''' Save the current status of the messages '''
        if self.conversion_func:
            self.conversion_func(self.origin, self.messages)
        else:
            self.origin.clear()
            self.origin += self.messages
        self.ignore_close_event = True
        self.close()
        self.close_inst = None

    def setSize(self,x_pos=0.4,y_pos=0.3,width=0.3,height=0.2):
        """Set the Window Size
        """
        # Set Window Size
        self.screen = self.app.primaryScreen()
        self.size = self.screen.size()
        self.setGeometry(
            int(self.size.width() * x_pos),
            int(self.size.height() * y_pos),
            int(self.size.width() * width),
            int(self.size.height() * height),
        )
        self.setCentralWidget(QWidget(self))

    def closeEvent(self, event) -> None:
        ''' Interrupt the close event, prompt the user if they want to save changes '''
        # Prompt user that we should save
        if self.ignore_close_event:
            return
        if not self.original == self.messages:
            buttons = QMessageBox.Save | QMessageBox.Discard | QMessageBox.Cancel
            reply = QMessageBox.question(self, 'Save Changes',
                                         "Do you want to save changes?",
                                         buttons,
                                         QMessageBox.Save)
            if reply == QMessageBox.Save:
                self.save()
            elif reply == QMessageBox.Discard:
                self.close_inst = None
            elif reply == QMessageBox.Cancel:
                event.ignore()

    def loadMessages(self, messages):
        ''' Load up the current messages provided '''
        list_widget = self.main_wgt.left_list
        # Message format is:
        # Assume these are defined as Msg Objects
        self.messages = messages
        list_widget.clear()
        for message in messages:
            list_widget.addItem(getname(message))

        list_widget.setCurrentRow(0)

    def msgChanged(self, item):
        ''' The user may have selected a different message to configure, 
        first ensure it is a legitimate item. '''
        if item:
            try:
                selected_msg = next(a for a in self.messages if getname(a) == item.text())
                self.displayMsg(selected_msg)
            except StopIteration:
                pass
        else:
            self.grid_wgt.msg_name.setText('')
            self.grid_wgt.msg_dest.setText('')
            self.grid_wgt.func.setText('')
            self.grid_wgt.maxwait.setText('')
            self.grid_wgt.mode.setCurrentText('ExpectsReply')
            self.param_config.left_list.clear()

    def paramChanged(self, item):
        ''' User has changed the parameter name '''
        list_widget = self.main_wgt.left_list
        list_item = list_widget.currentItem()
        try:
            if list_item:
                selected_msg = next(a for a in self.messages if getname(a) == list_item.text())
                if item:
                    key = next(a for a in list(selected_msg.parameters.objects.keys()) if
                               a == item.text())
                    self.displayParameter(key, selected_msg.parameters.objects[key])
                    return
            self.paramname.setText('')
            self.paramval.setText('')
        except StopIteration:
            pass

    def addParam(self):
        ''' User has added a parameter '''
        list_widget = self.param_config.left_list
        existing_names = [list_widget.item(index).text() for index in range(list_widget.count())]
        new_name = generateUniqueName(existing_names, 'func')
        try:
            msg_list_widget = self.main_wgt.left_list
            list_item = msg_list_widget.currentItem()
            selected_msg = next(a for a in self.messages if getname(a) == list_item.text())
            list_widget.addItem(new_name)
            selected_msg.parameters.objects[new_name] = '1'
            self.param_config.left_list.setCurrentRow(self.param_config.left_list.count()-1)
        except StopIteration:
            pass

    def remParam(self):
        ''' User has removed a parameter '''
        list_widget = self.main_wgt.left_list
        list_item = list_widget.currentItem()
        try:
            selected_msg = next(a for a in self.messages if getname(a) == list_item.text())
            list_widget = self.param_config.left_list
            del selected_msg.parameters.objects[list_widget.currentItem().text()]
            list_widget.takeItem(list_widget.row(list_widget.currentItem()))
        except StopIteration:
            pass

    def addMsg(self):
        ''' Add a new message '''
        list_widget = self.main_wgt.left_list
        existing_names = [list_widget.item(index).text() for index in range(list_widget.count())]
        new_name = generateUniqueName(existing_names, 'Message')
        new_msg = MARTe2Message('+' + new_name)
        new_msg.parameters = MARTe2ConfigurationDatabase(objects={})
        self.messages += [new_msg]
        list_widget.addItem(new_msg.configuration_name.replace('+', ''))
        list_widget.setCurrentRow(list_widget.count()-1)

    def displayMsg(self, msg):
        ''' Display the provided message in our QMessage widget '''
        self.grid_wgt.msg_name.setText(msg.configuration_name.replace('+', ''))
        self.grid_wgt.msg_dest.setText(msg.destination)
        self.grid_wgt.func.setText(msg.function)
        self.grid_wgt.maxwait.setText(str(msg.maxwait))
        self.grid_wgt.mode.setCurrentText(msg.mode)
        self.param_config.left_list.clear()
        for key, _ in msg.parameters.objects.items():
            self.param_config.left_list.addItem(key)
        self.param_config.left_list.setCurrentRow(0)


    def displayParameter(self, key, value):
        ''' Display the current parameter '''
        self.paramname.setText(key)
        self.paramval.setText(value)

    def remMsg(self):
        ''' Remove the currently selected message '''
        list_widget = self.main_wgt.left_list
        list_item = list_widget.currentItem()
        try:
            if list_item:
                selected_msg = next(a for a in self.messages if getname(a) == list_item.text())
                self.messages = [a for a in self.messages if not getname(a) == list_item.text()]
                del selected_msg
                list_widget.takeItem(list_widget.row(list_item))
        except StopIteration:
            pass

    def msgNameChg(self):
        ''' Change the message name '''
        list_widget = self.main_wgt.left_list
        list_item = list_widget.currentItem()
        try:
            if list_item:
                selected_msg = next(a for a in self.messages if getname(a) == list_item.text())
                selected_msg.configuration_name = '+' + self.grid_wgt.msg_name.text()
                list_item.setText(self.grid_wgt.msg_name.text())
        except StopIteration:
            pass

    def destChg(self):
        ''' Change to the destination of the message '''
        list_widget = self.main_wgt.left_list
        list_item = list_widget.currentItem()
        try:
            if list_item:
                selected_msg = next(a for a in self.messages if getname(a) == list_item.text())
                selected_msg.destination = self.grid_wgt.msg_dest.text()
        except StopIteration:
            pass

    def funcChg(self):
        ''' Change the function of the message '''
        list_widget = self.main_wgt.left_list
        list_item = list_widget.currentItem()
        try:
            if list_item:
                selected_msg = next(a for a in self.messages if getname(a) == list_item.text())
                selected_msg.function = self.grid_wgt.func.text()
        except StopIteration:
            pass

    def modeChg(self):
        ''' Change the mode of the message '''
        list_widget = self.main_wgt.left_list
        list_item = list_widget.currentItem()
        try:
            if list_item:
                selected_msg = next(a for a in self.messages if getname(a) == list_item.text())
                selected_msg.mode = self.grid_wgt.mode.currentText()
        except StopIteration:
            pass

    def waitChg(self):
        ''' Change the wait time of the message - note, -1 is infinite time '''
        list_widget = self.main_wgt.left_list
        list_item = list_widget.currentItem()
        try:
            if list_item:
                selected_msg = next(a for a in self.messages if getname(a) == list_item.text())
                selected_msg.maxwait = self.grid_wgt.maxwait.text()
        except StopIteration:
            pass

    def paramNameChg(self):
        ''' The parameter name has been changed '''
        list_widget = self.main_wgt.left_list
        list_item = list_widget.currentItem()
        try:
            if list_item:
                selected_msg = next(a for a in self.messages if getname(a) == list_item.text())
                list_widget = self.param_config.left_list.currentItem()
                if list_widget:
                    value = copy.copy(selected_msg.parameters.objects[list_widget.text()])
                    del selected_msg.parameters.objects[list_widget.text()]
                    selected_msg.parameters.objects[self.paramname.text()] = value
                    self.param_config.left_list.currentItem().setText(self.paramname.text())
        except StopIteration:
            pass

    def paramValChg(self):
        ''' Parameter value has been changed '''
        list_widget = self.main_wgt.left_list
        list_item = list_widget.currentItem()
        try:
            if list_item:
                selected_msg = next(a for a in self.messages if getname(a) == list_item.text())
                list_widget = self.param_config.left_list.currentItem()
                if list_widget:
                    selected_msg.parameters.objects[list_widget.text()] = self.paramval.text()
        except StopIteration:
            pass



'''
This GAM is used and useful for running a MARTe2 Configuration where you want it to end execution
after an expected number of iterations.

It requires a input signal but ignores this - it is purely needed because MARTe2 doesn't allow
GAMs without signals.

The parameter MaxCycles is what sets the number of iterations the application should run for.
'''
from martepy.marte2.gam import MARTe2GAM

class EndGAM(MARTe2GAM): #pylint: disable=W0223
    '''
    This GAM is used and useful for running a MARTe2 Configuration where you want
    it to end execution after an expected number of iterations.

    It requires a input signal but ignores this - it is purely needed because MARTe2 doesn't allow
    GAMs without signals.

    The parameter MaxCycles is what sets the number of iterations the application should run for.
    '''
    def __init__(self, #pylint: disable=W0102,W1113
                    configuration_name: str = 'End',
                    maxcycles=500,
                    input_signals=[],
                    *args, **kwargs
                ):
        #assert all(('Default' in d['MARTeConfig'] for n, d in output_signals))
        super().__init__(
                configuration_name = configuration_name,
                class_name = 'EndGAM',
                input_signals=input_signals,
                *args, **kwargs
            )
        self.maxcycles = maxcycles

    def writeGamConfig(self, config_writer):
        '''
        Write the MaxCycles parameter for our GAM
        '''
        config_writer.writeNode('MaxCycles', f'{self.maxcycles}')

def initialize(factory, plugin_datastore) -> None:
    ''' Register EndGAM to our given factory '''
    factory.registerBlock("EndGAM", EndGAM, plugin_datastore)



''' The basic config object definition from which all pythonic MARTe2 representations 
should derive. '''
import copy

import martepy.marte2.configwriting as marteconfig
from martepy.marte2.serialize import Serializable
from martepy.functions.extra_functions import normalizeSignal

class MARTe2ConfigObject(Serializable):
    """An object which will write configuration when called"""

    def __init__(self, configuration_name=''):
        super().__init__()
        self._config_writer = None
        self.configuration_name = configuration_name

    def __call__(self, config_writer=None):
        if not config_writer:
            config_writer = self._config_writer
        self.write(config_writer)

    def __str__(self):
        s = marteconfig.StringConfigWriter()
        self.write(s)
        return str(s)

    def __repr__(self):
        class_attributes = [attr for attr in dir(self) if
                            not callable(getattr(self, attr)) and not attr.startswith("__")]
        class_attributes.remove('id')
        string_content = ''
        for i in class_attributes:
            string_content += f'{i} = {getattr(self, i)}\n'
        return string_content

    def setWriter(self, config_writer):
        ''' Set the current writer '''
        self._config_writer = config_writer

    def setTab(self, tab):
        ''' Set the writers tab from within the object. '''
        if self._config_writer:
            self._config_writer.setTab(tab)
        else:
            raise TypeError('This MARTe1ConfigObject has no _config_writer')

    @staticmethod
    def writeSignals(defs, config_writer):
        """Helper function for writing out signals sections. The defs variable
        is a list of tuples of form (str, dict) where the string is a unique
        identifier for a signal and the dict should contain a key 'MARTeConfig'
        and may contain further keys with information about the signal. On
        'MARTeConfig' there should be a dict of key value pairs used to create
        leaf nodes in the signal configuration section.
        """
        for signal_name, details in defs:
            config_writer.startSection(signal_name)
            signal_details = normalizeSignal(copy.deepcopy(details))

            for key, value in signal_details['MARTeConfig'].items():
                config_writer.writeNode(key, value)
            config_writer.endSection(signal_name)

    def write(self, config_writer):
        """Write this configuration object into the provided config writer"""
        raise NotImplementedError

    def serialize(self):
        """Write this configuration object into the provided config writer"""
        return {'configuration_name': self.configuration_name}

    def deserialize(self, data, hashmap={}, restore_id=True, *args, **kwargs): # pylint: disable=W1113, W0613
        """Write this configuration object into the provided config writer"""
        self.configuration_name = data['configuration_name']
        return data



''' Pythonic representation of the HTTP Service object '''

from martepy.marte2.config_object import MARTe2ConfigObject
from martepy.functions.extra_functions import getname

class MARTe2HttpService(MARTe2ConfigObject):
    ''' Pythonic representation of the HTTP Service object '''
    def __init__(self,
                    configuration_name: str = 'WebRoot',
                    port = 8084,
                    webroot = "WebRoot",
                    timeout = 0,
                    listenmaxconnections = 255,
                    accepttimeout = 1000,
                    maxnumberofthreads = 8,
                    minnumberofthreads = 1,
                ):
        self.class_name = 'HttpService'
        super().__init__()
        self.configuration_name = configuration_name
        self.port = port
        self.webroot = webroot
        self.timeout = timeout
        self.listenmaxconnections = listenmaxconnections
        self.accepttimeout = accepttimeout
        self.maxnumberofthreads = maxnumberofthreads
        self.minnumberofthreads = minnumberofthreads

    def write(self, config_writer):
        ''' Write the configuration of our object '''
        config_writer.startClass('+' + getname(self), self.class_name)
        config_writer.writeNode('Port', f'"{self.port}"')
        config_writer.writeNode('WebRoot', f'"{self.webroot}"')
        config_writer.writeNode('Timeout', f'"{self.timeout}"')
        config_writer.writeNode('ListenMaxConnections', f'"{self.listenmaxconnections}"')
        config_writer.writeNode('AcceptTimeout', f'"{self.accepttimeout}"')
        config_writer.writeNode('MaxNumberOfThreads', f'"{self.maxnumberofthreads}"')
        config_writer.writeNode('MinNumberOfThreads', f'"{self.minnumberofthreads}"')
        config_writer.endSection('+' + getname(self))

    def serialize(self):
        ''' Serialize the object '''
        res = super().serialize()
        res['class_name'] = self.class_name
        res['port'] = self.port
        res['webroot'] = self.webroot
        res['timeout'] = self.timeout
        res['listenmaxconnections'] = self.listenmaxconnections
        res['accepttimeout'] = self.accepttimeout
        res['maxnumberofthreads'] = self.maxnumberofthreads
        res['minnumberofthreads'] = self.minnumberofthreads
        return res

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True, # pylint: disable=W0221
                    factory=None) -> bool:
        ''' Deserialize the given data to our class instance '''
        super().deserialize(data, hashmap, restore_id, factory)
        self.class_name = data["class_name"]
        self.port = data["port"]
        self.webroot = data["webroot"]
        self.timeout = data["timeout"]
        self.listenmaxconnections = data["listenmaxconnections"]
        self.accepttimeout = data["accepttimeout"]
        self.maxnumberofthreads = data["maxnumberofthreads"]
        self.minnumberofthreads = data["minnumberofthreads"]
        return self

def initialize(factory, plugin_datastore) -> None:
    ''' Register the object with the factory '''
    factory.registerBlock("HttpService", MARTe2HttpService, plugin_datastore)



''' Pythonic representation of the WaveformChirp GAM'''

from martepy.marte2.gam import MARTe2GAM
from martepy.marte2.qt_functions import (addInputSignalsSection,
                                         addOutputSignalsSection,
                                         addLineEdit)

class WaveformChirpGAM(MARTe2GAM):
    ''' Pythonic representation of the WaveformChirp GAM'''
    def __init__(self,
                    configuration_name: str = 'WaveChirp',
                    input_signals: list = [],
                    output_signals: list = [],
                    amplitude = 1.0,
                    frequency1 = 1.0,
                    frequency2 = 1.0,
                    phase = 0.0,
                    offset = 0.0,
                    starttriggertime = [],
                    stoptriggertime = []
                ):
        super().__init__(
                configuration_name = configuration_name,
                class_name = 'WaveformGAM::WaveformChirp',
                input_signals = input_signals,
                output_signals = output_signals,
            )
        self.amplitude = amplitude
        self.frequency1 = frequency1
        self.frequency2 = frequency2
        self.phase = phase
        self.offset = offset
        self.starttriggertime = starttriggertime
        self.stoptriggertime = stoptriggertime

    def writeGamConfig(self, config_writer):
        ''' Write our GAM Configuration '''
        def validateTime(name, value):
            if not (len(value) == 0 or value == '' or value == '{}'):
                if isinstance(value, list):
                    value = '{' + ' '.join([str(a) for a in value]) + '}'
                config_writer.writeNode(name,'{' + value.replace('{','').replace('}','') + '}')

        config_writer.writeNode("Amplitude",self.amplitude)
        config_writer.writeNode("Frequency1",self.frequency1)
        config_writer.writeNode("Frequency2",self.frequency2)
        config_writer.writeNode("Phase",self.phase)
        config_writer.writeNode("Offset",self.offset)
        validateTime("StartTriggerTime", self.starttriggertime)
        validateTime("StopTriggerTime", self.stoptriggertime)


    def write(self, config_writer):
        ''' Overwrite the writer so we can implement a time signal '''
        config_writer.startClass('+' + self.configuration_name.lstrip('+'), self.class_name)
        self.writeGamConfig(config_writer)
        if self.input_signals:
            config_writer.startSection('Time')
            self.writeSignals(self.input_signals, config_writer)
            config_writer.endSection('Time')
        if self.output_signals:
            config_writer.startSection('OutputSignals')
            self.writeSignals(self.output_signals, config_writer)
            config_writer.endSection('OutputSignals')
        config_writer.endSection('+' + self.configuration_name.lstrip('+'))

    def serialize(self):
        ''' Serialize the object '''
        res = super().serialize()
        res['parameters']['amplitude'] = self.amplitude
        res['parameters']['frequency1'] = self.frequency1
        res['parameters']['frequency2'] = self.frequency2
        res['parameters']['phase'] = self.phase
        res['parameters']['offset'] = self.offset
        res['parameters']['starttriggertime'] = self.starttriggertime
        res['parameters']['stoptriggertime'] = self.stoptriggertime
        return res

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True) -> bool:
        ''' Deserialize the given object to the class instance '''
        super().deserialize(data, hashmap, restore_id)
        # Now we build up
        self.amplitude = data['parameters']["amplitude"]
        self.frequency1 = data['parameters']["frequency1"]
        self.frequency2 = data['parameters']["frequency2"]
        self.phase = data['parameters']["phase"]
        self.offset = data['parameters']["offset"]
        self.starttriggertime = data['parameters']["starttriggertime"]
        self.stoptriggertime = data['parameters']["stoptriggertime"]
        return self

    @staticmethod
    def loadParameters(mainpanel_instance, node):
        """This function is intended to be for the GUI where it can
        call the static instance of the class directly to generate
        the appropriate parameter modifier for the node in XMARTe2.
        """

        def validateTime(field):
            if not (len(node.parameters[field]) == 0 or
                    node.parameters[field] == '' or
                    node.parameters[field]== '{}'):
                if isinstance(node.parameters[field], list):
                    times = [str(a) for a in node.parameters[field]]
                    node.parameters[field] = '{' + ' '.join(times) + '}'

        app_def = mainpanel_instance.parent.API.getServiceByName('ApplicationDefinition')
        datasource = app_def.configuration['misc']['gamsources'][0]
        addInputSignalsSection(mainpanel_instance, node, False)

        addOutputSignalsSection(mainpanel_instance, node, 3, False, datasource=datasource)

        addLineEdit(mainpanel_instance, node, "Amplitude: ", 'amplitude', 3, 0)

        addLineEdit(mainpanel_instance, node, "Frequency 1: ", 'frequency1', 3, 2)

        addLineEdit(mainpanel_instance, node, "Frequency 2: ", 'frequency2', 4, 0)

        addLineEdit(mainpanel_instance, node, "Phase: ", 'phase', 4, 2)

        addLineEdit(mainpanel_instance, node, "Offset: ", 'offset', 5, 0)

        validateTime('starttriggertime')
        validateTime('stoptriggertime')

        addLineEdit(mainpanel_instance, node, "Start Trigger Time: ", 'starttriggertime', 5, 2)

        addLineEdit(mainpanel_instance, node, "Stop Trigger Time: ", 'stoptriggertime', 6, 0)

def initialize(factory, plugin_datastore) -> None:
    ''' Initialize the objects with the factory '''
    factory.registerBlock("WaveformGAM::WaveformChirp", WaveformChirpGAM, plugin_datastore)
    factory.registerBlock("WaveformChirp", WaveformChirpGAM, plugin_datastore)
    factory.registerBlock("WaveformChirpGAM", WaveformChirpGAM, plugin_datastore)



''' Pythonic representation of the Message GAM'''

import copy

from PyQt5 import QtCore
from PyQt5.QtWidgets import (
    QGroupBox,
    QHBoxLayout,
    QListWidgetItem,
    QMainWindow,
    QPushButton,
    QWidget,
    QSizePolicy,
)
from qtpy.QtWidgets import QVBoxLayout

from martepy.marte2.gam import MARTe2GAM
from martepy.marte2.objects.message import MARTe2Message
from martepy.marte2.qt_classes import AddRemoveHBtn, PanelledListConfig, MessageConfigWindow
from martepy.marte2.objects.referencecontainer import MARTe2ReferenceContainer
from martepy.marte2.qt_functions import (addComboEdit,
                                         addInputSignalsSection,
                                         createComboEdit,
                                         createLineEdit,
                                         defineSaveCancelButtons,
                                         generateUniqueName,
                                         recursivelySetEnabled,
                                         setSize,
                                         showErrorDialog,
                                         textExistsInListWidget)
from martepy.marte2.objects.configuration_database import MARTe2ConfigurationDatabase
from martepy.functions.extra_functions import getname, findIndexByDictKey
from martepy.marte2.objects.statemachine.eventtrigger import MARTe2EventConditionTrigger

class MFactory(): # pylint: disable=R0903
    ''' Basic hard wired factory for the MessageGAMs sub components '''
    def create(self, name='EventConditionTrigger'):
        ''' Create function for building the class '''
        if name == 'EventConditionTrigger':
            return MARTe2EventConditionTrigger
        if name == 'ReferenceContainer':
            return MARTe2ReferenceContainer
        if name == 'ConfigurationDatabase':
            return MARTe2ConfigurationDatabase
        if name == 'Message':
            return MARTe2Message
        raise ValueError(f"""Object type not found in factory for MessageGAM,
 this is a bespoke factory class, check the message_gam.py. Name not found: {name}""")

class MessageGAM(MARTe2GAM):
    ''' Pythonic representation of the Message GAM'''
    def __init__(self,
                    configuration_name: str = 'MessageGAM',
                    input_signals: list = [],
                    output_signals: list = [],
                    triggeron: int = 1,
                    events = None,
                ):
        self.triggeron = triggeron
        self.mfactory = MFactory()
        self.events = events # Should be a reference container object
        if not self.events:
            self.events = MARTe2ReferenceContainer('+Events')

        #assert all(('Default' in d['MARTeConfig'] for n, d in output_signals))
        super().__init__(
                configuration_name = configuration_name,
                class_name = 'MessageGAM',
                input_signals = input_signals,
                output_signals = output_signals,
            )

    def writeGamConfig(self, config_writer):
        ''' Basic GAM Configuration Write up '''
        config_writer.writeNode("TriggerOnChange",self.triggeron)
        self.events.write(config_writer)

    def serialize(self):
        ''' Serialize the object '''
        res: dict = super().serialize()
        res['parameters']['triggeron'] = self.triggeron
        # Need to serialize this
        if self.events:
            res['parameters']['events'] = self.events.serialize()
        res['label'] = "MessageGAM"
        res['block_type'] = 'MessageGAM'
        res['class_name'] = 'MessageGAM'
        res['title'] = f"{self.configuration_name} (MessageGAM)"
        return res

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True) -> bool:
        ''' Deserialize the object '''
        super().deserialize(data, hashmap, restore_id)
        self.triggeron = data['parameters']["triggeron"]
        # Need to convert our events objects to messages and parameters
        self.events = MARTe2ReferenceContainer().deserialize(data['parameters']["events"],
                                                             factory=self.mfactory)
        return self

    @staticmethod
    def loadParameters(mainpanel_instance, node):
        """This function is intended to be for the GUI where it can call the
        static instance of the class directly to generate
        the appropriate parameter modifier for the node in XMARTe2.
        """
        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum)
        addInputSignalsSection(mainpanel_instance, node, False, False)
        mainpanel_instance.configbarBox.addWidget(spacer, 0, 2)

        addComboEdit(mainpanel_instance, node, 'Trigger on Change: ',
                     'triggeron', 2, 0, ['1','0'])

        def runWindow():
            mainpanel_instance.parent.newwindow = EventsWindow(mainpanel_instance.parent, node)
        config_btn = QPushButton("Configure Events")
        config_btn.clicked.connect(runWindow)
        mainpanel_instance.configbarBox.addWidget(config_btn, 2, 2)
        mainpanel_instance.configbarBox.addWidget(spacer, 2, 3)

class EventsWindow(QMainWindow):
    ''' Event Configuration window for the MessageGAM '''
    def __init__(self, parent=None, node=None):
        super().__init__(parent)
        self.node = node
        self.events = copy.deepcopy(self.node.parameters['events'])
        self.setWindowTitle(f"Events configuration for: {getname(self.node)}")

        self.msg_window = None
        setSize(self, self.parent().app, 0.225,0.25,0.55,0.5)

        self.main_wgt = PanelledListConfig(self, 0.25, 0.75)

        top_layout = self.main_wgt.v_layout
        self.setCentralWidget(self.main_wgt)

        # Organise left panel
        for event in self.events['objects']:
            # Need to get name from event and define how it is defined
            newitem = QListWidgetItem(event["configuration_name"].strip('+'))
            self.main_wgt.left_list.addItem(newitem)

        self.left_btns = AddRemoveHBtn()
        self.main_wgt.left_panel_vlayout.addWidget(self.left_btns)
        self.left_btns.add_btn.clicked.connect(self.addEvent)
        self.main_wgt.left_list.itemSelectionChanged.connect(self.selectedEventChanged)
        self.left_btns.remove_btn.clicked.connect(self.removeEvent)
        defineSaveCancelButtons(top_layout, self.save, self.cancel)

        # Now organise right hand panel
        self.eventname = createLineEdit(self.main_wgt.right_panel_vlayout,
                                        'Event Name:', self.eventNameChg)

        trigger_box = self.defineTriggerBox()

        self.msg_wgt = QWidget()
        hlayout = QHBoxLayout()
        self.msg_wgt.setLayout(hlayout)
        spacerb = QWidget()
        spacerb.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        hlayout.addWidget(spacerb)
        self.config_msgs = QPushButton("Configure Messages")
        self.config_msgs.clicked.connect(self.openMsgs)
        hlayout.addWidget(self.config_msgs)

        self.main_wgt.right_panel_vlayout.addWidget(trigger_box)
        self.main_wgt.right_panel_vlayout.addWidget(self.msg_wgt)

        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.main_wgt.right_panel_vlayout.addWidget(spacer)

        if len(self.events) == 0:
            recursivelySetEnabled(self.main_wgt.right_panel_vlayout, False)
        else:
            self.main_wgt.left_list.setCurrentRow(0)
        self.show()

    def defineTriggerBox(self):
        ''' Reduction of statements in init - builds the trigger box '''
        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        # Now add group box for defining the event trigger
        trigger_box = QGroupBox("Event Trigger:")
        trig_grp_layout = QVBoxLayout()
        trigger_box.setLayout(trig_grp_layout)

        self.trig_panel_wgt = PanelledListConfig(trigger_box, 0.3, 0.7)
        trig_grp_layout.addWidget(self.trig_panel_wgt)

        self.trig_btns = AddRemoveHBtn()
        self.trig_btns.add_btn.clicked.connect(self.addTrigger)
        self.trig_btns.remove_btn.clicked.connect(self.remTrigger)
        self.trig_panel_wgt.left_panel_vlayout.addWidget(self.trig_btns)
        signal_names = [a.label for a in self.node.inputs]
        self.trig_sig = createComboEdit(self.trig_panel_wgt.right_panel_vlayout,
                                        'Signal Name:', self.trigNameChg, signal_names)
        self.trig_val = createLineEdit(self.trig_panel_wgt.right_panel_vlayout,
                                       'Trigger Value:', self.trigValChg)
        self.trig_panel_wgt.left_list.itemSelectionChanged.connect(self.selectedTrigChanged)
        self.trig_panel_wgt.right_panel_vlayout.addWidget(spacer)
        return trigger_box

    def eventNameChg(self, value):
        ''' Event Name has changed '''
        selected_item = self.main_wgt.left_list.currentItem()
        event = selected_item.text()
        selected_item.setText(value)
        index = findIndexByDictKey(self.events['objects'], 'configuration_name', event)
        self.events['objects'][index]['configuration_name'] = '+' + value

    def openMsgs(self):
        ''' Open the Message Window '''
        selected_item = self.main_wgt.left_list.currentItem()
        if selected_item:
            event = selected_item.text()
            index = findIndexByDictKey(self.events['objects'], 'configuration_name', event)
            msgs = self.events["objects"][index]['objects']
            actual_msgs = [MARTe2Message().deserialize(a) for a in msgs]
            origin = self.events["objects"][index]['objects']
            self.msg_window = None
            self.msg_window = MessageConfigWindow(self.node.application,
                                                  self.node.application.app,
                                                  actual_msgs, origin,
                                                  f"Configure Event Messages for event {event}",
                                                  self.msg_window,
                                                  self.convert)

    def convert(self, original, msgs):
        ''' Convert the messages to the GUI acceptable format '''
        # Convert from MARTe2Message to a serialized method so we can
        # save this in .xmt format without changing external code in the GUI
        original.clear()
        original += [a.serialize() for a in msgs]

    def save(self):
        ''' Save our changes '''
        self.node.parameters['events'] = copy.deepcopy(self.events)
        self.close()

    def cancel(self):
        ''' Close without saving '''
        self.close()

    def trigNameChg(self, new_trig_idx):
        ''' Selected trigger has changed '''
        new_trig_name = self.trig_sig.itemText(new_trig_idx)
        found_items = self.trig_panel_wgt.left_list.findItems("*", QtCore.Qt.MatchWildcard)
        if new_trig_name not in [item.text() for item in found_items]:
            selected_item = self.main_wgt.left_list.currentItem()
            if selected_item:
                event = selected_item.text()
                index = findIndexByDictKey(self.events['objects'], 'configuration_name', event)
                trig_name = self.trig_panel_wgt.left_list.currentItem()
                if trig_name:
                    old_trig_name = trig_name.text()
                    try:
                        value = self.events['objects'][index]['eventtriggers'][old_trig_name]
                        del self.events['objects'][index]['eventtriggers'][old_trig_name]
                        self.events['objects'][index]['eventtriggers'][new_trig_name] = value
                        trig_name.setText(new_trig_name)
                    except (AttributeError, ValueError):
                        pass
        else:
            showErrorDialog("""This signal is already in use, you cannot
 have multiple triggers set to the same signal input.""")
            self.trig_sig.currentIndexChanged.disconnect()
            self.trig_sig.setCurrentText(self.trig_panel_wgt.left_list.currentItem().text())
            self.trig_sig.currentIndexChanged.connect(self.trigNameChg)

    def trigValChg(self, value):
        ''' Trigger value has changed '''
        selected_item = self.main_wgt.left_list.currentItem()
        if selected_item:
            event = selected_item.text()
            index = findIndexByDictKey(self.events['objects'], 'configuration_name', event)
            trig_name = self.trig_panel_wgt.left_list.currentItem()
            if trig_name:
                trig_name = trig_name.text()
                self.events['objects'][index]['eventtriggers'][trig_name] = value

    def addTrigger(self):
        ''' Add a trigger to our message '''
        event = self.eventname.text()
        index = findIndexByDictKey(self.events['objects'], 'configuration_name', event)
        if not self.events['objects'][index]['eventtriggers']:
            recursivelySetEnabled(self.trig_panel_wgt.right_panel_vlayout, True)
        try:
            # In future this should get the next unique label name -
            # for now, just gives the first one
            count = 0
            defaultsignal = None
            while count < len(self.node.inputs):
                if textExistsInListWidget(self.trig_panel_wgt.left_list,
                                          self.node.inputs[count].label):
                    count += 1
                    continue
                defaultsignal = self.node.inputs[count].label

            if defaultsignal is None:
                showErrorDialog("All available signals have defined triggers")
                return
        except (AttributeError, ValueError):
            showErrorDialog("""No signals available in the MessageGAM
 to select as the trigger source.""")
            return
        new_trigger = QListWidgetItem(defaultsignal)
        self.trig_panel_wgt.left_list.addItem(new_trigger)
        self.trig_panel_wgt.left_list.setCurrentItem(new_trigger)
        self.events['objects'][index]['eventtriggers'][defaultsignal] = '0'
        # Now set the lineedit values
        index = self.trig_sig.findText(defaultsignal)
        self.trig_sig.currentIndexChanged.disconnect()
        self.trig_sig.setCurrentIndex(index)
        self.trig_sig.currentIndexChanged.connect(self.trigNameChg)
        self.trig_val.setText('0')

    def remTrigger(self):
        ''' Remove the trigger '''
        event = self.eventname.text()
        index = findIndexByDictKey(self.events['objects'], 'configuration_name', event)
        selected_item = self.trig_panel_wgt.left_list.currentItem()
        if selected_item:
            del self.events['objects'][index]['eventtriggers'][selected_item.text()]
            currentRow = self.trig_panel_wgt.left_list.currentRow()
            self.trig_panel_wgt.left_list.takeItem(currentRow)
            if currentRow > 0:
                self.trig_panel_wgt.left_list.setCurrentRow(currentRow - 1)
            else:
                self.trig_panel_wgt.left_list.setCurrentRow(0)
        if not self.events['objects'][index]['eventtriggers']:
            recursivelySetEnabled(self.trig_panel_wgt.right_panel_vlayout, False)

    def addEvent(self):
        ''' Add the event '''
        if len(self.events) == 0:
            recursivelySetEnabled(self.main_wgt.right_panel_vlayout, True)
        newevent = self.defaultEvent()
        current_item = QListWidgetItem(newevent['configuration_name'].lstrip('+'))
        self.main_wgt.left_list.addItem(current_item)
        self.main_wgt.left_list.setCurrentItem(current_item)
        self.eventname.textChanged.disconnect()
        self.eventname.setText(newevent['configuration_name'].lstrip('+'))
        self.eventname.textChanged.connect(self.eventNameChg)
        self.events['objects'] += [newevent]
        recursivelySetEnabled(self.trig_panel_wgt.right_panel_vlayout, False)

    def selectedEventChanged(self):
        ''' User has selected a different event case '''
        # We can probably reuse the event change state easily here
        # by activating the selected trig change and selected msg change
        # if those lists have length, select the first index.
        selected_item = self.main_wgt.left_list.currentItem()
        if selected_item:
            event = selected_item.text()
            index = findIndexByDictKey(self.events['objects'], 'configuration_name', event)
            self.eventname.textChanged.disconnect()
            self.eventname.setText(event)
            self.eventname.textChanged.connect(self.eventNameChg)
            # Remove all previous items from left lists for trig and msg
            self.trig_panel_wgt.left_list.clear()
            if index != -1:
                trig_names = list(self.events['objects'][index]['eventtriggers'])
                self.trig_panel_wgt.left_list.addItems(trig_names)
                self.trig_panel_wgt.left_list.update()
                # Now activate those
                if len(trig_names) > 0:
                    self.trig_panel_wgt.left_list.setCurrentRow(0)
                    recursivelySetEnabled(self.trig_panel_wgt.right_panel_vlayout, True)
                else:
                    recursivelySetEnabled(self.trig_panel_wgt.right_panel_vlayout, False)

    def selectedTrigChanged(self):
        ''' User has changed the trigger selected '''
        selected_item = self.main_wgt.left_list.currentItem()
        if selected_item:
            # Hotfix for badly displayed error during population
            self.trig_sig.currentIndexChanged.disconnect()
            event = selected_item.text()
            index = findIndexByDictKey(self.events['objects'], 'configuration_name', event)
            trig_name = self.trig_panel_wgt.left_list.currentItem()
            try:
                if trig_name:
                    trig_name = trig_name.text()
                    index = self.trig_sig.findText(trig_name)
                    self.trig_sig.setCurrentIndex(index)
                    self.trig_val.setText(self.events['objects'][index]['eventtriggers'][trig_name])
            except (AttributeError, ValueError):
                pass
            self.trig_sig.currentIndexChanged.connect(self.trigNameChg)

    def defaultEvent(self):
        ''' Return a default event configuration '''
        existing_names = [d['configuration_name'].lstrip('+') for d in self.events['objects']]
        new_name = generateUniqueName(existing_names, 'NewEvent')
        return MARTe2EventConditionTrigger('+' + new_name, eventtriggers={}, msgs=[]).serialize()

    def removeEvent(self):
        ''' Remove the current event '''
        selected_item = self.main_wgt.left_list.currentItem()

        def isObj(d):
            return d['configuration_name'].lstrip('+') != selected_item.text()

        if selected_item:
            self.events['objects'] = [d for d in self.events['objects'] if isObj(d)]
            currentRow = self.main_wgt.left_list.currentRow()
            self.main_wgt.left_list.takeItem(currentRow)
            if currentRow > 0:
                self.main_wgt.left_list.setCurrentRow(currentRow-1)
            else:
                self.main_wgt.left_list.setCurrentRow(0)
            if len(self.events) == 0:
                recursivelySetEnabled(self.main_wgt.right_panel_vlayout, False)


def initialize(factory, plugin_datastore) -> None:
    ''' Initialize the object with the factory '''
    factory.registerBlock("MessageGAM", MessageGAM, plugin_datastore)



''' Import SDN Subscriber and Publisher all in one go '''

from martepy.marte2.datasources.sdn.subscriber import SDNSubscriber
from martepy.marte2.datasources.sdn.publisher import SDNPublisher

__all__ = [
        'SDNSubscriber',
        'SDNPublisher',
]



"""extra_functions.py contains additional useful functions"""


class ConfigGeneratorError(Exception):
    """Generic exception for errors encountered during configuration generation"""

def getname(object_cls):
    ''' Return the name of an object, omitting the + '''
    return object_cls.configuration_name.lstrip('+')

def normalizeSignal(signal):
    ''' Checks that a signal has elements and dimensions and removes these if not
    otherwise it ensures they are both set. It also handles defaults in this context,
    if a default is with zero dimensions/elements, it will remove the {} encapsulation
    of default.'''
    if 'NumberOfDimensions' in signal['MARTeConfig']:
        if 'NumberOfElements' in signal['MARTeConfig']:
            dimensions = signal['MARTeConfig']['NumberOfDimensions']
            elements = signal['MARTeConfig']['NumberOfElements']
            if int(dimensions) == 1 and int(elements) == 1:
                del signal['MARTeConfig']['NumberOfDimensions']
                del signal['MARTeConfig']['NumberOfElements']

    if 'NumberOfElements' not in signal['MARTeConfig'].keys():
        if 'Default' in signal['MARTeConfig'].keys():
            current_default = str(signal['MARTeConfig']['Default'])
            signal['MARTeConfig']['Default'] = current_default.lstrip('{').rstrip('}')

    return signal

def findIndexByDictKey(list_i, key, value):
    ''' This will search through a list of dictionaries and find a key value matching pair
    # this is intended for where dictionaries are identified by a unique key/value.'''
    for index, dictionary in enumerate(list_i):
        if dictionary.get(key).lstrip('+') == value:
            return index
    return -1

def isint(value):
    ''' Is our value a valid integer? '''
    try:
        int(value)
        return True
    except ValueError:
        return False

def isfloat(value):
    ''' Is our value a valid float value? '''
    try:
        float(value)
        return True
    except ValueError:
        return False

def calculateStackSize(signals):
    """:param signals: inputs and outputs
    :type signals: list[tuple[str, dict[]]]

    :return stacksize: total size of variable types
    :rtype: int

    Given a function, how big is the input and output signal stack size of the variables
    """
    types = {"int8": 1, "int16": 2, "int32": 4, "int64": 8,
             "uint8": 1, "uint16": 2, "uint32": 4, "uint64": 8,
             "float32": 4, "float64": 8}
    stacksize = 0
    for signal in signals:
        signal_type = signal[1]["MARTeConfig"]["Type"]
        num = 1
        if 'NumberOfElements' in signal[1]["MARTeConfig"].keys():
            num = int(signal[1]["MARTeConfig"]['NumberOfElements'])
        stacksize = stacksize + types[signal_type] * num
    return stacksize

def form(sig_name, type_, num=None, datasource='DDB0', alias=None, default=None):
    """:param str sig_name: signal name
    :param str type_: signal type
    :param int num: number of elements
    :param str datasource: MARTe datasource
    :param str alias: signal alias

    :return: output signals formatted
    :rtype: list[tuple[str, dict[]]]

    Formats the input/output signals
    """
    if alias is None:
        alias = sig_name
    formatted = [(sig_name, {'MARTeConfig': {'DataSource': datasource,
                                             'Type': type_,
                                             'Alias': alias}})]
    if default:
        formatted[0][1]['MARTeConfig']['Default'] = default
    if num:
        formatted[0][1]['MARTeConfig']['NumberOfElements'] = num
    return formatted



''' Pythonic representation of the UDP Sender '''

from martepy.marte2.datasource import MARTe2DataSource
from martepy.marte2.qt_functions import addComboEdit, addLineEdit, addInputSignalsSection

class UDPSender(MARTe2DataSource):
    ''' Pythonic representation of the UDP Sender '''
    def __init__(self,
                    configuration_name: str = 'UDPSender',
                    input_signals = [],
                    output_signals = [],
                    port: int = 0,
                    cpumask: int = 0xFFFFFFFF,
                    executionmode = 'IndependentThread',
                    address = '127.0.0.1',
                    numberofpretriggers = 0,
                    numberofposttriggers = 0,
                    stacksize = 10000000
                ):
        super().__init__(
                configuration_name = configuration_name,
                class_name = 'UDPDrv::UDPSender',
                input_signals = input_signals
            )
        self.port = port
        self.cpumask = cpumask
        self.executionmode = executionmode
        self.address = address
        self.numberofpretriggers = numberofpretriggers
        self.numberofposttriggers = numberofposttriggers
        self.stacksize = stacksize

    def writeDatasourceConfig(self, config_writer):
        ''' Write the Datasource configuration '''
        config_writer.writeNode('Port', f'{self.port}')
        config_writer.writeNode('Address', f'"{self.address}"')
        config_writer.writeNode('ExecutionMode', f'{self.executionmode}')
        config_writer.writeNode('NumberOfPreTriggers', f'{self.numberofpretriggers}')
        config_writer.writeNode('NumberOfPostTriggers', f'{self.numberofposttriggers}')
        if isinstance(self.cpumask, str):
            if 'x' in self.cpumask:
                config_writer.writeNode('CPUMask', self.cpumask)
            else:
                config_writer.writeNode('CPUMask', hex(int(self.cpumask)))
        else:
            config_writer.writeNode('CPUMask', hex(self.cpumask))
        config_writer.writeNode('StackSize', f'{self.stacksize}')
        self.writeInputSignals(config_writer, section_name='Signals')

    def serialize(self):
        ''' Serialize the object '''
        res: dict = super().serialize()
        res['parameters']['port'] = self.port
        res['parameters']['cpumask'] = self.cpumask
        res['parameters']['numberofpretriggers'] = self.numberofpretriggers
        res['parameters']['numberofposttriggers'] = self.numberofposttriggers
        res['parameters']['stacksize'] = self.stacksize
        res['parameters']['address'] = self.address
        res['parameters']['executionmode'] = self.executionmode
        res['label'] = "UDPSender"
        res['block_type'] = 'UDPSender'
        res['class_name'] = 'UDPSender'
        res['title'] = f"{self.configuration_name} (UDPSender)"
        return res

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True) -> bool:
        ''' Deserialize the object to the class instance '''
        res = super().deserialize(data, hashmap, restore_id)
        self.port = data['parameters']["port"]
        self.cpumask = data['parameters']["cpumask"]
        self.address = data['parameters']["address"]
        self.executionmode = data['parameters']["executionmode"]
        self.numberofpretriggers = data['parameters']["numberofpretriggers"]
        self.numberofposttriggers = data['parameters']["numberofposttriggers"]
        self.stacksize = data['parameters']["stacksize"]
        return res

    @staticmethod
    def loadParameters(mainpanel_instance, node):
        """This function is intended to be for the GUI where it can
        call the static instance of the class directly to generate
        the appropriate parameter modifier for the node in XMARTe2.
        """
        addInputSignalsSection(mainpanel_instance,node,False, datasource=node.configuration_name)

        addLineEdit(mainpanel_instance, node, "Port: ", 'port', 2, 0)

        addLineEdit(mainpanel_instance, node, "Address: ", 'address', 2, 2)

        addComboEdit(mainpanel_instance, node, "Execution Mode: ", 'executionmode',
                     3, 0, ['IndependentThread', 'RealTimeThread'])

        addLineEdit(mainpanel_instance, node, "Number of Pretriggers: ",
                    'numberofpretriggers', 3, 2)

        addLineEdit(mainpanel_instance, node, "Number of Posttriggers: ",
                    'numberofposttriggers', 4, 0)

        if isinstance(node.parameters['cpumask'], str):
            if 'x' in node.parameters['cpumask']:
                node.parameters['cpumask'] = node.parameters['cpumask']
            else:
                node.parameters['cpumask'] = hex(int(node.parameters['cpumask']))
        else:
            node.parameters['cpumask'] = hex(node.parameters['cpumask'])

        addLineEdit(mainpanel_instance, node, "CPU Mask: ", 'cpumask', 4, 2)

        addLineEdit(mainpanel_instance, node, "Stack Size: ", 'stacksize', 5, 0)

def initialize(factory, plugin_datastore) -> None:
    ''' Initialize the object with the factory '''
    factory.registerBlock("UDPSender", UDPSender, plugin_datastore)



''' Pythonic representative class of the File Reader Datasource ''' 

from functools import partial

from PyQt5.QtWidgets import (
    QPushButton,
    QWidget,
    QLineEdit,
    QLabel,
    QHBoxLayout,
    QFileDialog
)

from martepy.marte2.datasource import MARTe2DataSource
from martepy.marte2.qt_functions import (addComboEdit,
                                         addOutputSignalsSection,
                                         addLineEdit,
                                         paraChange)

class FileReader(MARTe2DataSource):
    ''' Pythonic representation of the FileReader '''
    def __init__(self,
                    configuration_name: str = 'FileReader',
                    input_signals = [],
                    output_signals = [],
                    filename = "",
                    fileformat = "csv",
                    separator = ",",
                    interpolate = "no",
                    eof = "Rewind",
                    preload = "yes"
                ):
        super().__init__(
                configuration_name = configuration_name,
                class_name = 'FileDataSource::FileReader',
                output_signals = output_signals,
            )
        self.filename = filename
        self.fileformat = fileformat
        self.csvseparator = separator
        self.interpolate = interpolate
        self.eof = eof
        self.preload = preload

    def writeDatasourceConfig(self, config_writer):
        ''' Write the datasource configuration '''
        config_writer.writeNode('Filename', f'"{self.filename}"')
        config_writer.writeNode('FileFormat', f'"{self.fileformat}"')
        config_writer.writeNode('CSVSeparator', f'"{self.csvseparator}"')
        config_writer.writeNode('Interpolate', f'"{self.interpolate}"')
        config_writer.writeNode('EOF', f'"{self.eof}"')
        config_writer.writeNode('Preload', f'"{self.preload}"')
        self.writeOutputSignals(config_writer, section_name='Signals')

    def serialize(self):
        ''' Serialize our current configuration '''
        res: dict = super().serialize()
        res['parameters']['filename'] = self.filename
        res['parameters']['fileformat'] = self.fileformat
        res['parameters']['interpolate'] = self.interpolate
        res['parameters']['csvseparator'] = self.csvseparator
        res['parameters']['eof'] = self.eof
        res['parameters']['preload'] = self.preload

        # Override name
        res['label'] = "FileReader"
        res['block_type'] = 'FileReader'
        res['class_name'] = 'FileReader'
        res['title'] = f"{self.configuration_name} (FileReader)"
        return res

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True) -> bool:
        ''' Deserialize the object to our class instance '''
        res = super().deserialize(data, hashmap, restore_id)
        self.filename = data['parameters']["filename"]
        self.fileformat = data['parameters']["fileformat"]
        self.interpolate = data['parameters']["interpolate"]
        self.csvseparator = data['parameters']["csvseparator"]
        self.eof = data['parameters']["eof"]
        self.preload = data['parameters']["preload"]
        return res

    def writeSignals(self, defs, config_writer):
        """Use MARTe1ConfigObject.writeSignals but remove any DataSources."""
        omit_keys = ('DataSource','Alias')
        signals = [(
                n,
                dict(d, **{'MARTeConfig': {
                        k: v for k, v in d['MARTeConfig'].items() if k not in omit_keys
                    }})
            ) for n, d in defs]

        super().writeSignals(signals, config_writer)

    @staticmethod
    def loadParameters(mainpanel_instance, node):
        """This function is intended to be for the GUI where it can
        call the static instance of the class directly to generate
        the appropriate parameter modifier for the node in XMARTe2.
        """
        addOutputSignalsSection(mainpanel_instance, node, False)

        # Special file input
        wgt_label = QLabel("File name: ")
        wgt_field = QWidget()
        hlay = QHBoxLayout()
        wgt_field.setLayout(hlay)
        wgt_line = QLineEdit(mainpanel_instance)
        wgt_button = QPushButton("Browse")
        hlay.addWidget(wgt_line)
        hlay.addWidget(wgt_button)
        def filedialog():
            # Open a file dialog to select a file
            file_dialog = QFileDialog()
            file_dialog.setFileMode(QFileDialog.ExistingFile)
            if file_dialog.exec_():
                file_paths = file_dialog.selectedFiles()
                if file_paths:
                    # Read the text content from the selected file
                    file_path = file_paths[0]
                    wgt_line.setText(file_path)

        wgt_button.clicked.connect(filedialog)
        wgt_line.setText(str(node.parameters['filename']))
        wgt_line.textChanged.connect(partial(paraChange, wgt_line, node, 'filename'))
        mainpanel_instance.configbarBox.addWidget(wgt_label, 2, 0)
        mainpanel_instance.configbarBox.addWidget(wgt_field, 2, 1)

        addComboEdit(mainpanel_instance, node,"File format: ", 'fileformat',
                     2, 2, ['csv', 'binary'])

        addComboEdit(mainpanel_instance, node,"Interpolate: ", 'interpolate', 3, 0, ['yes', 'no'])

        addLineEdit(mainpanel_instance, node,"Separator: ",'csvseparator', 3, 2)

        addComboEdit(mainpanel_instance, node,"Preload: ", 'preload', 4, 0, ['yes', 'no'])

        addComboEdit(mainpanel_instance, node,"EOF: ", 'eof', 4, 2, ['Rewind', 'Last', 'Error'])

def initialize(factory, plugin_datastore) -> None:
    ''' Initialize the object with the factory '''
    factory.registerBlock("FileReader", FileReader, plugin_datastore)



''' Import all the available file datasources in one go '''

from martepy.marte2.datasources.files.writer import FileWriter
from martepy.marte2.datasources.files.reader import FileReader
from martepy.marte2.datasources.files.file_datasources import RFileReader, RFileWriter

__all__ = [
        'FileWriter',
        'FileReader',
        'RFileReader',
        'RFileWriter'
]



''' Pythonic representation of the Thread object '''

from martepy.marte2.config_object import MARTe2ConfigObject
from martepy.functions.extra_functions import getname

class NonDuplicatingList(list):
    ''' Ensure our list is non duplicating '''
    def append(self, item):
        ''' Ensure the append is not appending again '''
        if not any(item == list_item for list_item in self):
            super().append(item)

    def __iadd__(self, other):
        for item in other:
            self.append(item)
        return self

class MARTe2RealTimeThread(MARTe2ConfigObject):
    """Object for configuring RealTimeThreads for MARTe2 applications"""

    def __init__(self,
                    configuration_name: str = 'Thread',
                    cpu_mask: int = 0xFFFFFFFF,
                    functions: list = [],
                ):
        super().__init__()
        self.configuration_name = configuration_name
        self.class_name = 'RealTimeThread'
        self.cpu_mask = cpu_mask
        self.functions = functions

    def write(self, config_writer):
        ''' Write our configuration of this class '''
        config_writer.startClass('+' + getname(self), self.class_name)
        if isinstance(self.cpu_mask, str):
            if 'x' in self.cpu_mask:
                # Capitalize the letters only
                capitalized_hex = self.cpu_mask[:2] + self.cpu_mask[2:].upper()
                config_writer.writeNode('CPUs', capitalized_hex)
            else:
                hex_value = hex(int(self.cpu_mask))
                # Capitalize the letters only
                capitalized_hex = hex_value[:2] + hex_value[2:].upper()
                config_writer.writeNode('CPUs', capitalized_hex)
        else:
            hex_value = hex(self.cpu_mask)
            # Capitalize the letters only
            capitalized_hex = hex_value[:2] + hex_value[2:].upper()
            config_writer.writeNode('CPUs', capitalized_hex)
        function_names = NonDuplicatingList()
        [function_names.append(getname(i)) for i in self.functions] # pylint: disable=W0106
        config_writer.writeMARTe2Vector('Functions', function_names, formatAsFloat = False)
        config_writer.endSection('+' + getname(self))

    def serialize(self):
        ''' Serialize our object '''
        res = super().serialize()
        res['class_name'] = self.class_name
        res['cpu_mask'] = self.cpu_mask
        if len(self.functions) > 0:
            res['functions'] = [a.serialize() for a in self.functions]
        else:
            res['functions'] = []
        return res

    def __eq__(self, other):
        attributes = ['configuration_name', 'class_name', 'cpu_mask']
        if isinstance(other, self.__class__):
            for attribute in attributes:
                if not getattr(self, attribute) == getattr(other, attribute):
                    return False
            for function, ofunction in zip(self.functions, other.functions):
                if not function == ofunction:
                    return False
            return True
        return False

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True, # pylint: disable=W0221
                    factory=None) -> bool:
        ''' Deserialize our data into an instance of our class '''

        def createObj(a):
            return factory.create(a['class_name'])().deserialize(a)

        super().deserialize(data, hashmap, restore_id, factory=factory)
        self.class_name = data["class_name"]
        self.cpu_mask = data["cpu_mask"]
        self.functions = [createObj(a) for a in data["functions"]]
        return self

def initialize(factory, plugin_datastore) -> None:
    ''' Register us with the factory '''
    factory.registerBlock("RealTimeThread", MARTe2RealTimeThread, plugin_datastore)



''' Pythonic representation of the HTTP message interface '''

from martepy.marte2.config_object import MARTe2ConfigObject
from martepy.functions.extra_functions import getname
from martepy.marte2.objects.message import MARTe2Message

class MARTe2HttpMessageInterface(MARTe2ConfigObject):
    ''' Pythonic representation of the HTTP message interface '''
    def __init__(self,
                    configuration_name: str = 'WebRoot',
                    objects = [],
                ):
        self.class_name = 'HttpMessageInterface'
        super().__init__()
        self.configuration_name = configuration_name
        self.objects = objects

    def write(self, config_writer):
        ''' Write the configuration of our object '''
        config_writer.startClass('+' + getname(self), self.class_name)
        for obj in self.objects:
            obj.write(config_writer)
        config_writer.endSection('+' + getname(self))

    def serialize(self):
        ''' Serialize the object '''
        res = super().serialize()
        res['class_name'] = self.class_name
        res["objects"] = [a.serialize() for a in self.objects]
        return res

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True, # pylint: disable=W0221
                    factory=None) -> bool:
        ''' Deserialize the given data to our class instance '''
        super().deserialize(data, hashmap, restore_id, factory=factory)
        self.class_name = data["class_name"]
        self.objects = [MARTe2Message().deserialize(a) for a in data["objects"]]
        return self

def initialize(factory, plugin_datastore) -> None:
    ''' Register the object with the factory '''
    factory.registerBlock("HttpMessageInterface", MARTe2HttpMessageInterface, plugin_datastore)



''' Pythonic representation of the WaveformPoints GAM'''

from martepy.marte2.gam import MARTe2GAM
from martepy.marte2.qt_functions import (addInputSignalsSection,
                                         addOutputSignalsSection,
                                         addLineEdit)

class WaveformPointsGAM(MARTe2GAM):
    ''' Pythonic representation of the WaveformPoints GAM'''
    def __init__(self,
                    configuration_name: str = 'WavePoints',
                    input_signals: list = [],
                    output_signals: list = [],
                    points = [],
                    times = [],
                    starttriggertime = [],
                    stoptriggertime = []
                ):
        super().__init__(
                configuration_name = configuration_name,
                class_name = 'WaveformGAM::WaveformPointsDef',
                input_signals = input_signals,
                output_signals = output_signals,
            )
        self.points = points
        self.times = times
        self.starttriggertime = starttriggertime
        self.stoptriggertime = stoptriggertime

    def writeGamConfig(self, config_writer):
        ''' Write our GAM Configuration '''
        def validateTime(name, value):
            if not (len(value) == 0 or value == '' or value == '{}'):
                if isinstance(value, list):
                    value = '{' + ' '.join([str(a) for a in value]) + '}'
                config_writer.writeNode(name,'{' + value.replace('{','').replace('}','') + '}')

        if isinstance(self.points, list):
            self.points = '{' + ' '.join([str(a) for a in self.points]) + '}'
        config_writer.writeNode("Points",'{' + self.points.replace('{','').replace('}','') + '}')
        if isinstance(self.times, list):
            self.times = '{' + ' '.join([str(a) for a in self.times]) + '}'
        config_writer.writeNode("Times",'{' + self.times.replace('{','').replace('}','') + '}')
        validateTime("StartTriggerTime", self.starttriggertime)
        validateTime("StopTriggerTime", self.stoptriggertime)

    def serialize(self):
        ''' Serialize the object '''
        res = super().serialize()
        res['parameters']['points'] = self.points
        res['parameters']['times'] = self.times
        res['parameters']['starttriggertime'] = self.starttriggertime
        res['parameters']['stoptriggertime'] = self.stoptriggertime
        return res

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True) -> bool:
        ''' Deserialize the given object to the class instance '''
        super().deserialize(data, hashmap, restore_id)
        # Now we build up
        self.points = data['parameters']["points"]
        self.times = data['parameters']["times"]
        self.starttriggertime = data['parameters']["starttriggertime"]
        self.stoptriggertime = data['parameters']["stoptriggertime"]
        return self

    @staticmethod
    def loadParameters(mainpanel_instance, node):
        """This function is intended to be for the GUI where it can
        call the static instance of the class directly to generate
        the appropriate parameter modifier for the node in XMARTe2.
        """

        def validateTime(field):
            if not (len(node.parameters[field]) == 0 or
                    node.parameters[field] == '' or
                    node.parameters[field]== '{}'):
                if isinstance(node.parameters[field], list):
                    times = [str(a) for a in node.parameters[field]]
                    node.parameters[field] = '{' + ' '.join(times) + '}'

        app_def = mainpanel_instance.parent.API.getServiceByName('ApplicationDefinition')
        datasource = app_def.configuration['misc']['gamsources'][0]
        addInputSignalsSection(mainpanel_instance, node, False)

        addOutputSignalsSection(mainpanel_instance, node, 3, False, datasource=datasource)

        validateTime('points')
        validateTime('times')

        addLineEdit(mainpanel_instance, node, "Points: ", 'points', 3, 0)

        addLineEdit(mainpanel_instance, node, "Times: ", 'times', 3, 2)

        validateTime('starttriggertime')
        validateTime('stoptriggertime')

        addLineEdit(mainpanel_instance, node, "Start Trigger Time: ", 'starttriggertime', 4, 0)

        addLineEdit(mainpanel_instance, node, "Stop Trigger Time: ", 'stoptriggertime', 4, 2)

def initialize(factory, plugin_datastore) -> None:
    ''' Initialize the objects with the factory '''
    factory.registerBlock("WaveformGAM::WaveformPointsDef", WaveformPointsGAM, plugin_datastore)
    factory.registerBlock("WaveformPointsDef", WaveformPointsGAM, plugin_datastore)
    factory.registerBlock("WaveformPointsGAM", WaveformPointsGAM, plugin_datastore)



''' Pythonic representation of the IO GAM'''

from martepy.marte2.gam import MARTe2GAM
from martepy.marte2.qt_functions import addInputSignalsSection, addOutputSignalsSection

class IOGAM(MARTe2GAM):
    ''' Pythonic representation of the IO GAM'''
    def __init__(self,
                    configuration_name: str = 'IO',
                    input_signals: list = [],
                    output_signals: list = [],
                ):
        super().__init__(
                configuration_name = configuration_name,
                class_name = 'IOGAM',
                input_signals = input_signals,
                output_signals = output_signals,
            )

    def writeGamConfig(self, _):
        ''' Nothing to write for a IO GAM '''

    def serialize(self):
        ''' Serialize the object '''
        res = super().serialize()
        return res

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True) -> bool:
        ''' Deserialize the given object to the class instance '''
        super().deserialize(data, hashmap, restore_id)
        return self

    @staticmethod
    def loadParameters(mainpanel_instance, node):
        """This function is intended to be for the GUI where it can
        call the static instance of the class directly to generate
        the appropriate parameter modifier for the node in XMARTe2.
        """
        app_def = mainpanel_instance.parent.API.getServiceByName('ApplicationDefinition')
        datasource = app_def.configuration['misc']['gamsources'][0]
        addInputSignalsSection(mainpanel_instance, node, False)

        addOutputSignalsSection(mainpanel_instance, node, 3, False, datasource=datasource)

def initialize(factory, plugin_datastore) -> None:
    ''' Initialize the object with the factory '''
    factory.registerBlock("IOGAM", IOGAM, plugin_datastore)



''' Pythonic representation of Message '''

from martepy.marte2.config_object import MARTe2ConfigObject
from martepy.marte2.objects.configuration_database import MARTe2ConfigurationDatabase

class MARTe2Message(MARTe2ConfigObject):
    ''' Pythonic representation of Message '''
    def __init__(self,
                    configuration_name: str = 'Message',
                    destination = "App",
                    function = "StopCurrentStateExecution",
                    parameters = MARTe2ConfigurationDatabase(),
                    mode = "ExpectsReply",
                    maxwait = 0,
                ):
        super().__init__(
            )
        self.class_name = "Message"
        self.configuration_name = configuration_name
        self.destination = destination
        self.function = function
        self.parameters = parameters
        self.mode = mode
        self.maxwait = maxwait

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            comparisons = ['configuration_name', 'destination','function',
                           'parameters','mode','maxwait']
            for item in comparisons:
                try:
                    if not getattr(self, item) == getattr(other, item):
                        return False
                except AttributeError:
                    return False
            return True
        return False

    def write(self, config_writer):
        ''' Write our configuration of this class '''
        def sanitize(value):
            return str(value).replace('"','')
        config_writer.startClass('+' + self.configuration_name.lstrip('+'),"Message")
        dest = sanitize(self.destination).replace('$','')
        config_writer.writeNode('Destination', f'"{dest}"')
        config_writer.writeNode('Function', f'"{sanitize(self.function)}"')
        config_writer.writeNode('MaxWait', f'"{sanitize(self.maxwait)}"')
        if self.mode != "":
            config_writer.writeNode('Mode', f'"{sanitize(self.mode)}"')
        if self.parameters:
            if self.parameters.objects:
                self.parameters.write(config_writer)
        config_writer.endSection('+' + self.configuration_name.lstrip('+'))

    def serialize(self):
        ''' Serialize our object '''
        res = super().serialize()
        res['class_name'] = self.class_name
        res['destination'] = self.destination
        res['maxwait'] = self.maxwait
        res['function'] = self.function
        res['parameters'] = self.parameters.serialize()
        res['mode'] = self.mode
        return res

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True, # pylint: disable=W0221
                    factory=None) -> bool:
        ''' Deserialize our data into an instance of our class '''
        super().deserialize(data, hashmap, restore_id, factory=factory)
        self.class_name = data["class_name"]
        self.destination = data["destination"]
        self.function = data["function"]
        self.maxwait = data['maxwait']

        self.parameters = MARTe2ConfigurationDatabase().deserialize(data["parameters"],
                                                                    factory=factory)
        self.mode = data["mode"]
        return self

def initialize(factory, plugin_datastore) -> None:
    ''' Register us with the factory '''
    factory.registerBlock("Message", MARTe2Message, plugin_datastore)



''' Pythonic representation of configuration database '''

from martepy.marte2.config_object import MARTe2ConfigObject
from martepy.functions.extra_functions import isint, isfloat

class MARTe2ConfigurationDatabase(MARTe2ConfigObject):
    ''' Pythonic representation of configuration database '''
    def __init__(self,
                    configuration_name: str = '+Parameters',
                    class_name = "ConfigurationDatabase",
                    objects = {},
                ):
        super().__init__()
        self.class_name = class_name
        self.objects = objects
        self.configuration_name = configuration_name

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            if self.configuration_name == other.configuration_name:
                if self.objects == other.objects:
                    return True
        return False

    def write(self, config_writer):
        ''' Write the objects we are holding '''
        config_writer.startClass('+' + self.configuration_name.lstrip('+'), self.class_name)
        for key, value in self.objects.items():
            linevalue = ''
            if isint(value) or isfloat(value):
                linevalue = f"{value}"
            else:
                strline = value.replace('"','').replace("'",'')
                linevalue = f'"{strline}"'
            config_writer.writeNode(key,linevalue)
        config_writer.endSection('+' + self.configuration_name.lstrip('+'))

    def serialize(self):
        ''' Serialize our objects '''
        res = super().serialize()
        res['class_name'] = "ConfigurationDatabase"
        res["objects"] = self.objects
        return res

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True, # pylint: disable=W0221
                    factory=None) -> bool:
        ''' Deserialize our objects with a factory '''
        super().deserialize(data, hashmap, restore_id, factory=factory)
        self.class_name = "ConfigurationDatabase"
        # This part is tricky as a ref container can
        # contain any object so the factory needs to recircle here
        self.objects = data["objects"]
        for _, value in self.objects.items():
            if not (isint(value) or isfloat(value)):
                value = value.replace('"','').replace("'",'')
        return self

def initialize(factory, plugin_datastore) -> None:
    ''' Register us with the factory '''
    factory.registerBlock("ConfigurationDatabase", MARTe2ConfigurationDatabase, plugin_datastore)
